@@ -3,7 +3,7 @@
  *
  *	This file contains basic support for handling addresses.
  *
- * TkRat software and its included text is Copyright 1996-2000 by
+ * TkRat software and its included text is Copyright 1996-2002 by
  * Martin Forssn
  *
  * The full text of the legal notice is contained in the file called
@@ -34,6 +34,15 @@ typedef struct {
  */
 Tcl_HashTable aliasTable;
 
+/*
+ * This struct us ised when aliases are expanded
+ */
+typedef struct {
+    char *host;
+    int lookup_in_passwd;
+    int level;
+} AliasExpand;
+
 /*
  * The number of address entities created. This is used to create new
  * unique command names.
@@ -45,6 +54,17 @@ static int numAddresses = 0;
  */
 static unsigned long aliasMark = 0;
 
+/*
+ * States used while parsing addresses
+ */
+typedef enum {
+    STATE_NORMAL,
+    STATE_ESCAPED,
+    STATE_COMMENT,
+    STATE_QUOTED,
+    STATE_LITERAL
+} ParseState;
+
 /*
  * Internal functions
  */
@@ -54,6 +74,10 @@ static int AddressClean(Tcl_Obj *aPtr);
 static char *mem_store;
 #endif /* MEM_DEBUG */
 
+static int RatAddressIsMeRole(Tcl_Interp *interp, ADDRESS *adrPtr, char *role);
+static void RatExpandAlias(Tcl_Interp *interp, Tcl_DString *list,
+			   AliasExpand *ea);
+
 
 /*
  *----------------------------------------------------------------------
@@ -140,16 +164,17 @@ RatCreateAddressCmd(ClientData clientData, Tcl_Interp *interp, int objc,
 	Tcl_Obj *CONST objv[])
 {
     ADDRESS *adrPtr = NULL;
-    char *s;
+    char *s, *ch;
 
-    if (objc != 2) {
+    if (objc != 3) {
 	Tcl_AppendResult(interp, "wrong # args: should be \"",
-		Tcl_GetString(objv[0]), " address\"", (char *) NULL);
+		Tcl_GetString(objv[0]), " address role\"", (char *) NULL);
 	return TCL_ERROR;
     }
 
+    ch = RatGetCurrent(interp, RAT_HOST, Tcl_GetString(objv[2]));
     s = cpystr(Tcl_GetString(objv[1]));
-    rfc822_parse_adrlist(&adrPtr, s, currentHost);
+    rfc822_parse_adrlist(&adrPtr, s, ch);
     ckfree(s);
     RatEncodeAddresses(interp, adrPtr);
     RatInitAddresses(interp, adrPtr);
@@ -180,11 +205,13 @@ void
 RatInitAddresses(Tcl_Interp *interp, ADDRESS *addressPtr)
 {
     ADDRESS *adrPtr, *newPtr;
-    Tcl_DString result;
     char name[32];
+    Tcl_Obj *rPtr;
 
-    Tcl_DStringInit(&result);
-    Tcl_DStringGetResult(interp, &result);
+    rPtr = Tcl_GetObjResult(interp);
+    if (Tcl_IsShared(rPtr)) {
+	rPtr = Tcl_DuplicateObj(rPtr);
+    }
     for (adrPtr = addressPtr; adrPtr; adrPtr = adrPtr->next) {
 	newPtr = mail_newaddr();
 	if (adrPtr->personal)	{
@@ -196,11 +223,11 @@ RatInitAddresses(Tcl_Interp *interp, ADDRESS *addressPtr)
 	if (adrPtr->host)	newPtr->host = cpystr(adrPtr->host);
 	if (adrPtr->error)	newPtr->error = cpystr(adrPtr->error);
 	sprintf(name, "RatAddress%d", numAddresses++);
-	Tcl_CreateCommand(interp, name, RatAddress, (ClientData) newPtr,
+	Tcl_CreateObjCommand(interp, name, RatAddress, (ClientData) newPtr,
 		RatDeleteAddress);
-	Tcl_DStringAppendElement(&result, name);
+	Tcl_ListObjAppendElement(interp, rPtr, Tcl_NewStringObj(name, -1));
     }
-    Tcl_DStringResult(interp, &result);
+    Tcl_SetObjResult(interp, rPtr);
 }
 
 /*
@@ -222,74 +249,49 @@ RatInitAddresses(Tcl_Interp *interp, ADDRESS *addressPtr)
  */
 
 int
-RatAddress(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
+RatAddress(ClientData clientData, Tcl_Interp *interp, int objc,
+	   Tcl_Obj *const objv[])
 {
     ADDRESS *adrPtr = (ADDRESS*)clientData;
+    Tcl_CmdInfo info;
     Tcl_Obj *oPtr;
     int useup;
 
-    if (argc < 2) {
-	Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
-		" option ?arg?\"", (char *) NULL);
-	return TCL_ERROR;
-    }
-    if (!strcmp(argv[1], "isMe")) {
-	if (3 == argc) {
-	    Tcl_GetBoolean(interp, argv[2], &useup);
+    if (objc < 2) goto usage;
+    if (!strcmp(Tcl_GetString(objv[1]), "isMe")) {
+	if (3 == objc) {
+	    Tcl_GetBooleanFromObj(interp, objv[2], &useup);
 	} else {
 	    useup = 1;
 	}
-	if (RatAddressIsMe(interp, adrPtr, useup)) {
-	    Tcl_SetResult(interp, "1", TCL_STATIC);
-	} else {
-	    Tcl_SetResult(interp, "0", TCL_STATIC);
-	}
+	Tcl_SetObjResult(interp, Tcl_NewBooleanObj(
+	    RatAddressIsMe(interp, adrPtr, useup)));
 	return TCL_OK;
 	
-    } else if (!strcmp(argv[1], "compare")) {
-	Tcl_CmdInfo info;
-	if (argc != 3) {
-	    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
-		    " compare address\"", (char *) NULL);
-	    return TCL_ERROR;
-	}
-	if (0 == Tcl_GetCommandInfo(interp, argv[2], &info)) {
+    } else if (!strcmp(Tcl_GetString(objv[1]), "compare")) {
+	if (objc != 3) goto usage;
+	if (0 == Tcl_GetCommandInfo(interp, Tcl_GetString(objv[2]), &info)) {
 	    Tcl_AppendResult(interp, "there is no address entity \"",
-		    argv[2], "\"", (char *) NULL);
+		    Tcl_GetString(objv[2]), "\"", (char *) NULL);
 	    return TCL_ERROR;
 	}
-	if (RatAddressCompare(adrPtr, (ADDRESS*)info.clientData)) {
-	    Tcl_SetResult(interp, "1", TCL_STATIC);
-	} else {
-	    Tcl_SetResult(interp, "0", TCL_STATIC);
-	}
+	Tcl_SetObjResult(interp, Tcl_NewBooleanObj(
+	    RatAddressCompare(adrPtr, (ADDRESS*)info.objClientData)));
 	return TCL_OK;
 
-    } else if (!strcmp(argv[1], "set")) {
-	if (argc != 5) {
-	    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
-		    " set personal name host\"", (char *) NULL);
-	    return TCL_ERROR;
-	}
+    } else if (!strcmp(Tcl_GetString(objv[1]), "set")) {
+	if (objc != 5) goto usage;
 	ckfree(adrPtr->mailbox);
-	if (adrPtr->personal) {
-	    ckfree(adrPtr->personal);
-	}
-	if (adrPtr->host) {
-	    ckfree(adrPtr->host);
-	}
-	adrPtr->personal = cpystr(argv[2]);
-	adrPtr->mailbox = cpystr(argv[3]);
-	adrPtr->host = cpystr(argv[4]);
+	ckfree(adrPtr->personal);
+	ckfree(adrPtr->host);
+	adrPtr->personal = cpystr(Tcl_GetString(objv[2]));
+	adrPtr->mailbox = cpystr(Tcl_GetString(objv[3]));
+	adrPtr->host = cpystr(Tcl_GetString(objv[4]));
 	return TCL_OK;
 
-    } else if (!strcmp(argv[1], "get")) {
-	if (argc != 3) {
-	    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
-		    " get form\"", (char *) NULL);
-	    return TCL_ERROR;
-	}
-	if (!strcasecmp(argv[2], "rfc822")) {
+    } else if (!strcmp(Tcl_GetString(objv[1]), "get")) {
+	if (objc != 3) goto usage;
+	if (!strcasecmp(Tcl_GetString(objv[2]), "rfc822")) {
 	    if (adrPtr->personal) {
 		char *personal;
 
@@ -301,32 +303,30 @@ RatAddress(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
 		Tcl_AppendToObj(oPtr, RatAddressMail(adrPtr), -1);
 		Tcl_AppendToObj(oPtr, ">", 1);
 		Tcl_SetObjResult(interp, oPtr);
-		ckfree(personal);
 	    } else {
 		Tcl_SetResult(interp, RatAddressMail(adrPtr), TCL_VOLATILE);
 	    }
 	    return TCL_OK;
 
-	} else if (!strcasecmp(argv[2], "mail")) {
+	} else if (!strcmp(Tcl_GetString(objv[2]), "mail")) {
 	    Tcl_SetResult(interp, RatAddressMail(adrPtr), TCL_VOLATILE);
 	    return TCL_OK;
 
-	} else if (!strcasecmp(argv[2], "name")) {
+	} else if (!strcmp(Tcl_GetString(objv[2]), "name")) {
 	    if (adrPtr->personal) {
 		Tcl_SetResult(interp, adrPtr->personal, TCL_VOLATILE);
 	    }
 	    return TCL_OK;
 
 	} else {
-	    Tcl_AppendResult(interp, "bad form \"", argv[2],
-		    "\": must be one of rfc822, mail or name", (char *) NULL);
-	    return TCL_ERROR;
+	    goto usage;
 	}
-    } else {
-	Tcl_AppendResult(interp, "bad option \"", argv[1],
-		"\": must be one of isMe, compare, set or get", (char *) NULL);
-	return TCL_ERROR;
     }
+
+ usage:
+    Tcl_AppendResult(interp, "Illegal usage of \"", Tcl_GetString(objv[0]),
+		     "\"", (char *) NULL);
+    return TCL_ERROR;
 }
 
 /*
@@ -350,14 +350,51 @@ void
 RatDeleteAddress(ClientData clientData)
 {
     ADDRESS *adrPtr = (ADDRESS*)clientData;
-    if (adrPtr->personal)	ckfree(adrPtr->personal);
-    if (adrPtr->adl)		ckfree(adrPtr->adl);
-    if (adrPtr->mailbox)	ckfree(adrPtr->mailbox);
-    if (adrPtr->host)		ckfree(adrPtr->host);
-    if (adrPtr->error)		ckfree(adrPtr->error);
+    ckfree(adrPtr->personal);
+    ckfree(adrPtr->adl);
+    ckfree(adrPtr->mailbox);
+    ckfree(adrPtr->host);
+    ckfree(adrPtr->error);
     ckfree(adrPtr);
 }
 
+/*
+ *----------------------------------------------------------------------
+ *
+ * RatAddressIsMeInRole --
+ *
+ *      Checks if the address points to me in the given role
+ *
+ * Results:
+ *	If it is then non zero is returned otherwise zero.
+ *
+ * Side effects:
+ *	None.
+ *
+ *
+ *----------------------------------------------------------------------
+ */
+static int
+RatAddressIsMeRole(Tcl_Interp *interp, ADDRESS *adrPtr, char *role)
+{
+    char *from, *host;
+    ADDRESS *a = NULL;
+
+    host = cpystr(RatGetCurrent(interp, RAT_HOST, role));
+    from = cpystr(RatGetCurrent(interp, RAT_MAILBOX, role));
+    rfc822_parse_adrlist(&a, from, host);
+    ckfree(from);
+    ckfree(host);
+    if (a && adrPtr->mailbox && adrPtr->host
+	&& !strcasecmp(a->mailbox, adrPtr->mailbox)
+	&& !strcasecmp(a->host, adrPtr->host)) {
+	mail_free_address(&a);
+	return 1;
+    }
+    mail_free_address(&a);
+    return 0;
+}
+
 /*
  *----------------------------------------------------------------------
  *
@@ -369,40 +406,35 @@ RatDeleteAddress(ClientData clientData)
  *	If it is then non zero is returned otherwise zero.
  *
  * Side effects:
- *	May initialize the current* variables
+ *	None.
  *
  *
  *----------------------------------------------------------------------
  */
 
 int
-RatAddressIsMe(Tcl_Interp *interp, ADDRESS *adrPtr, int useUP)
+RatAddressIsMe(Tcl_Interp *interp, ADDRESS *adrPtr, int trustUser)
 {
-    char *from;
+    Tcl_Obj **objv, *oPtr;
+    Tcl_CmdInfo cmdInfo;
+    int objc, i;
 
     if (adrPtr == NULL) {
 	return 0;
     }
-    if (adrPtr->mailbox && !strcasecmp(adrPtr->mailbox, currentMailboxName)
-	    && (adrPtr->host && !strcasecmp(adrPtr->host, currentHost))) {
+    
+    if (RatAddressIsMeRole(interp, adrPtr, "")) {
 	return 1;
     }
-    if (NULL != (from = Tcl_GetVar2(interp,"option","from",TCL_GLOBAL_ONLY))
-	    && *from != '\0') {
-	ADDRESS *a = NULL;
-	char *s = cpystr(from);
-	rfc822_parse_adrlist(&a, s, currentHost);
-	ckfree(s);
-	if (a && adrPtr->mailbox && adrPtr->host
-		&& !strcasecmp(a->mailbox, adrPtr->mailbox)
-		&& !strcasecmp(a->host, adrPtr->host)) {
-	    mail_free_address(&a);
-	    return 1;
+
+    if (trustUser) {
+	oPtr = Tcl_GetVar2Ex(interp, "option", "roles", TCL_GLOBAL_ONLY);
+	Tcl_ListObjGetElements(interp, oPtr, &objc, &objv);
+	for (i=0; i<objc; i++) {
+	    if (RatAddressIsMeRole(interp, adrPtr, Tcl_GetString(objv[i]))) {
+		return 1;
+	    }
 	}
-	mail_free_address(&a);
-    }
-    if (useUP) {
-	Tcl_CmdInfo cmdInfo;
 	if (Tcl_GetCommandInfo(interp, "RatUP_IsMe", &cmdInfo)) {
 	    Tcl_DString cmd;
 	    int isMe;
@@ -412,7 +444,8 @@ RatAddressIsMe(Tcl_Interp *interp, ADDRESS *adrPtr, int useUP)
 	    Tcl_DStringAppendElement(&cmd, "RatUP_IsMe");
 	    Tcl_DStringAppendElement(&cmd,adrPtr->mailbox?adrPtr->mailbox:"");
 	    Tcl_DStringAppendElement(&cmd,adrPtr->host?adrPtr->host:"");
-	    Tcl_DStringAppendElement(&cmd,adrPtr->personal?adrPtr->personal:"");
+	    Tcl_DStringAppendElement(&cmd,
+				     adrPtr->personal ? adrPtr->personal : "");
 	    Tcl_DStringAppendElement(&cmd,adrPtr->adl?adrPtr->adl:"");
 	    if (TCL_OK == Tcl_Eval(interp, Tcl_DStringValue(&cmd))
 	    	    && (oPtr = Tcl_GetObjResult(interp))
@@ -437,7 +470,7 @@ RatAddressIsMe(Tcl_Interp *interp, ADDRESS *adrPtr, int useUP)
  *	If they are then zero is returned otherwise non zero.
  *
  * Side effects:
- *	May initialize the current* variables
+ *	None.
  *
  *
  *----------------------------------------------------------------------
@@ -511,9 +544,7 @@ RatAddressTranslate(Tcl_Interp *interp, ADDRESS *adrPtr)
 	    s = Tcl_GetString(oPtr);
 	    if (   (*s && (!(*destPtrPtr) || strcmp(s,*destPtrPtr)))
 		|| (!*s && *destPtrPtr)) {
-		if (*destPtrPtr) {
-		    ckfree(*destPtrPtr);
-		}
+		ckfree(*destPtrPtr);
 		if (*s) {
 		    *destPtrPtr = cpystr(s);
 		} else {
@@ -543,8 +574,10 @@ RatAddressTranslate(Tcl_Interp *interp, ADDRESS *adrPtr)
  */
 
 int
-RatAliasCmd(ClientData dummy, Tcl_Interp *interp,int objc,Tcl_Obj *CONST objv[])
+RatAliasCmd(ClientData dummy,Tcl_Interp *interp,int objc,Tcl_Obj *CONST objv[])
 {
+    Tcl_Obj *oPtr;
+
     if (objc < 2) {
 	Tcl_AppendResult(interp, "wrong # args: should be \"",
 		Tcl_GetString(objv[0]), " option ?arg?\"", (char *) NULL);
@@ -644,14 +677,13 @@ RatAliasCmd(ClientData dummy, Tcl_Interp *interp,int objc,Tcl_Obj *CONST objv[])
     } else if (!strcmp(Tcl_GetString(objv[1]), "get")) {
 	Tcl_HashEntry *entryPtr;
 	AliasInfo *aliasPtr;
-	Tcl_Obj *oPtr;
 
 	if (objc != 3) {
 	    Tcl_AppendResult(interp, "wrong # args: should be \"",
 		    Tcl_GetString(objv[0]), " get alias\"", (char *) NULL);
 	    return TCL_ERROR;
 	}
-	if (!(entryPtr=Tcl_FindHashEntry(&aliasTable, Tcl_GetString(objv[2])))){
+	if (!(entryPtr=Tcl_FindHashEntry(&aliasTable,Tcl_GetString(objv[2])))){
 	    Tcl_SetResult(interp, "Illegal alias", TCL_STATIC);
 	    return TCL_ERROR;
 	}
@@ -674,7 +706,6 @@ RatAliasCmd(ClientData dummy, Tcl_Interp *interp,int objc,Tcl_Obj *CONST objv[])
 	Tcl_HashEntry *entryPtr;
 	Tcl_HashSearch search;
 	AliasInfo *aliasPtr;
-	Tcl_Obj *oPtr;
 
 	if (objc != 3) {
 	    Tcl_AppendResult(interp, "wrong # args: should be \"",
@@ -703,7 +734,6 @@ RatAliasCmd(ClientData dummy, Tcl_Interp *interp,int objc,Tcl_Obj *CONST objv[])
 	
     } else if (!strcmp(Tcl_GetString(objv[1]), "read")) {
 	Tcl_Channel channel;
-	Tcl_Obj *oPtr = Tcl_NewObj();
 
 	if (objc != 3) {
 	    Tcl_AppendResult(interp, "wrong # args: should be \"",
@@ -715,9 +745,11 @@ RatAliasCmd(ClientData dummy, Tcl_Interp *interp,int objc,Tcl_Obj *CONST objv[])
 	    return TCL_ERROR;
 	}
 	Tcl_SetChannelOption(interp, channel, "-encoding", "utf-8");
+	oPtr = Tcl_NewObj();
 	while (0 <= Tcl_GetsObj(channel, oPtr) && !Tcl_Eof(channel)) {
 	    Tcl_AppendToObj(oPtr, ";", 1);
 	}
+        Tcl_Close(interp, channel);
 	return Tcl_EvalObjEx(interp, oPtr,  TCL_EVAL_DIRECT);
 	
     } else if (!strcmp(Tcl_GetString(objv[1]), "save")) {
@@ -730,12 +762,13 @@ RatAliasCmd(ClientData dummy, Tcl_Interp *interp,int objc,Tcl_Obj *CONST objv[])
 
 	if (objc != 4) {
 	    Tcl_AppendResult(interp, "wrong # args: should be \"",
-		    Tcl_GetString(objv[0])," save book filename\"",(char*)NULL);
+			     Tcl_GetString(objv[0])," save book filename\"",
+			     (char*)NULL);
 	    return TCL_ERROR;
 	}
 
-	Tcl_GetInt(interp, Tcl_GetVar2(interp, "option", "permissions",
-		TCL_GLOBAL_ONLY), &perm);
+	oPtr = Tcl_GetVar2Ex(interp, "option", "permissions", TCL_GLOBAL_ONLY);
+	Tcl_GetIntFromObj(interp, oPtr, &perm);
 	if (NULL == (channel = Tcl_OpenFileChannel(interp,
 		Tcl_GetString(objv[3]), "w", perm))) {
 	    return TCL_ERROR;
@@ -744,13 +777,13 @@ RatAliasCmd(ClientData dummy, Tcl_Interp *interp,int objc,Tcl_Obj *CONST objv[])
 	for (entryPtr = Tcl_FirstHashEntry(&aliasTable, &search);
 		entryPtr; entryPtr = Tcl_NextHashEntry(&search)) {
 	    aliasPtr = (AliasInfo*) Tcl_GetHashValue(entryPtr);
-	    if (strcmp(Tcl_GetString(objv[2]), Tcl_GetString(aliasPtr->book))) {
+	    if (strcmp(Tcl_GetString(objv[2]), Tcl_GetString(aliasPtr->book))){
 		continue;
 	    }
 	    lPtr = Tcl_NewObj();
 	    Tcl_ListObjAppendElement(interp, lPtr, aliasPtr->book);
 	    Tcl_ListObjAppendElement(interp, lPtr,
-		    Tcl_NewStringObj(Tcl_GetHashKey(&aliasTable, entryPtr),-1));
+		    Tcl_NewStringObj(Tcl_GetHashKey(&aliasTable,entryPtr),-1));
 	    Tcl_ListObjAppendElement(interp, lPtr, aliasPtr->fullname);
 	    Tcl_ListObjAppendElement(interp, lPtr, aliasPtr->content);
 	    Tcl_ListObjAppendElement(interp, lPtr, aliasPtr->comment);
@@ -769,32 +802,62 @@ RatAliasCmd(ClientData dummy, Tcl_Interp *interp,int objc,Tcl_Obj *CONST objv[])
 	
     } else if (!strcmp(Tcl_GetString(objv[1]), "expand1")
 	    || !strcmp(Tcl_GetString(objv[1]), "expand2")) {
-	ADDRESS *adrPtr, *baseAdrPtr = NULL;
-	struct passwd *pwPtr;
 	Tcl_HashEntry *entryPtr;
-	Tcl_DString newList;
 	AliasInfo *aliasPtr;
-	char buf[1024], *sPtr, *cPtr, *userPtr;
-	int level, rescan, useTable, inGroup, length;
+	AliasExpand ae;
+	Tcl_DString list;
+	char *role, *c, *s;
+	ADDRESS *adrPtr, *baseAdrPtr = NULL;
 
-	if (objc != 3) {
+	if (objc != 4) {
 	    Tcl_AppendResult(interp, "wrong # args: should be \"",
-		    Tcl_GetString(objv[0]), " expand[12] adrlist\"",
+		    Tcl_GetString(objv[0]), " expand[12] adrlist role\"",
 		    (char *) NULL);
 	    return TCL_ERROR;
 	}
 
+	/*
+	 * Check syntax
+	 */
+	s = cpystr(Tcl_GetString(objv[2]));
+	for (c=s; *c; c++) {
+	    if ('\n' == *c || '\r' == *c || '\t' == *c) {
+		*c = ' ';
+	    }
+	}
+	rfc822_parse_adrlist(&baseAdrPtr, s, "");
+	ckfree(s);
+	for (adrPtr = baseAdrPtr; adrPtr; adrPtr = adrPtr->next) {
+	    if (adrPtr->error || (adrPtr->host && adrPtr->host[0] == '.')) {
+		mail_free_address(&baseAdrPtr);
+		Tcl_SetResult(interp, "Error in address list", TCL_STATIC);
+		return TCL_ERROR;
+	    }
+	}
+	mail_free_address(&baseAdrPtr);
+
+
 	/*
 	 * Ignore empty addresses
 	 */
-	for (sPtr = Tcl_GetString(objv[2]);
-		    *sPtr && isspace((unsigned char)*sPtr); sPtr++);
-	if (!*sPtr) {
+	for (c = Tcl_GetString(objv[2]); *c && isspace((unsigned char)*c);c++);
+	if (!*c) {
 	    return TCL_OK;
 	}
+	
+	role = Tcl_GetString(objv[3]);
+	ae.host = RatGetCurrent(interp, RAT_HOST, role);
+	oPtr = Tcl_GetVar2Ex(interp, "option", "lookup_name", TCL_GLOBAL_ONLY);
+	Tcl_GetBooleanFromObj(interp, oPtr, &ae.lookup_in_passwd);
+	if (!strcmp(Tcl_GetString(objv[1]), "expand2")) {
+	    ae.level = 2;
+	} else {
+	    Tcl_GetIntFromObj(interp, Tcl_GetVar2Ex(interp, "option",
+		    "alias_expand", TCL_GLOBAL_ONLY), &ae.level);
+	}
 
 	/*
-	 * Create unique mark
+	 * Create unique mark and possibly mark all aliases
 	 */
 	if (0 == ++aliasMark) {
 	    Tcl_HashSearch search;
@@ -807,131 +870,11 @@ RatAliasCmd(ClientData dummy, Tcl_Interp *interp,int objc,Tcl_Obj *CONST objv[])
 	    }
 	}
 
-	Tcl_GetBoolean(interp, Tcl_GetVar2(interp, "option", "lookup_name",
-			               TCL_GLOBAL_ONLY), &useTable);
-
-	if (!strcmp(Tcl_GetString(objv[1]), "expand2")) {
-	    level = 2;
-	} else {
-	    Tcl_GetIntFromObj(interp, Tcl_GetVar2Ex(interp, "option",
-		    "alias_expand", TCL_GLOBAL_ONLY), &level);
-	}
-	Tcl_DStringInit(&newList);
-	Tcl_DStringAppend(&newList, sPtr, -1);
-	for (cPtr = Tcl_DStringValue(&newList); *cPtr; cPtr++) {
-	    if (isspace((unsigned char)*cPtr)) {
-		*cPtr = ' ';
-	    }
-	}
-	do {
-	    inGroup = rescan = 0;
-	    rfc822_parse_adrlist(&baseAdrPtr, Tcl_DStringValue(&newList), "");
-	    Tcl_DStringSetLength(&newList, 0);
-	    for (adrPtr = baseAdrPtr; adrPtr; adrPtr = adrPtr->next) {
-		if (!adrPtr->mailbox) {
-		    if (inGroup) {
-			Tcl_DStringAppend(&newList, ";", 1);
-			inGroup = 0;
-		    }
-		    if (adrPtr->next && adrPtr->next->mailbox) {
-			Tcl_DStringAppend(&newList, ", ", 2);
-		    }
-		    continue;
-		}
-		if (!inGroup && !adrPtr->host) {
-		    Tcl_DStringAppend(&newList, adrPtr->mailbox, -1);
-		    Tcl_DStringAppend(&newList, ":", 1);
-		    inGroup = 1;
-		    continue;
-		}
-		/*
-		 * Check if this matches an alias. If the host is an empty
-		 * string we just try the mailbox name as key, otherwise we
-		 * try mailbox@host.
-		 */
-		if (adrPtr->error || (adrPtr->host && adrPtr->host[0] == '.')){
-		    Tcl_DStringFree(&newList);
-		    if (adrPtr->error) {
-			Tcl_SetResult(interp, adrPtr->error, TCL_VOLATILE);
-		    } else {
-			Tcl_SetResult(interp, adrPtr->host, TCL_VOLATILE);
-		    }
-		    mail_free_address(&baseAdrPtr);
-		    return TCL_ERROR;
-		}
-		entryPtr = NULL;
-		userPtr = NULL;
-		if (adrPtr->host && *adrPtr->host) {
-		    snprintf(buf, sizeof(buf), "%s@%s",
-			    adrPtr->mailbox, adrPtr->host);
-		} else {
-		    RatStrNCpy(buf, adrPtr->mailbox, sizeof(buf));
-		}
-		if (!(entryPtr = Tcl_FindHashEntry(&aliasTable, buf))
-			&& useTable) {
-		    if ((pwPtr = getpwnam(buf))) {
-			RatStrNCpy(buf, pwPtr->pw_gecos, sizeof(buf));
-			if ((cPtr = strchr(buf, ','))) {
-			    *cPtr = '\0';
-			}
-			userPtr = buf;
-		    }
-		}
-		if (entryPtr) {
-		    aliasPtr = (AliasInfo*) Tcl_GetHashValue(entryPtr);
-		    if (aliasPtr->mark == aliasMark) {
-			continue;
-		    }
-		    switch (level) {
-		    case 0:
-			Tcl_DStringAppend(&newList, buf, -1);
-			break;
-		    case 1:
-			Tcl_DStringAppend(&newList, RatAddressMail(adrPtr), -1);
-			cPtr = Tcl_GetString(aliasPtr->fullname);
-			if (cPtr) {
-			    sprintf(buf, " (%.1020s)", cPtr);
-			    Tcl_DStringAppend(&newList, buf, strlen(buf));
-			}
-			break;
-		    case 2:
-			rescan = 1;
-			cPtr = Tcl_GetStringFromObj(aliasPtr->content, &length);
-			Tcl_DStringAppend(&newList, cPtr, length);
-			if (!(aliasPtr->flags & ALIAS_FLAG_ISLIST)
-				&& !(aliasPtr->flags & ALIAS_FLAG_NOFULLNAME)) {
-			    Tcl_DStringAppend(&newList, " (", 2);
-			    cPtr = Tcl_GetStringFromObj(aliasPtr->fullname,
-				    &length);
-			    Tcl_DStringAppend(&newList, cPtr, length);
-			    Tcl_DStringAppend(&newList, ")", 1);
-			}
-			break;
-		    }
-		    aliasPtr->mark = aliasMark;
-		} else if ((!adrPtr->host || !*adrPtr->host) &&
-			   !adrPtr->personal && level > 0 && userPtr) {
-		    Tcl_DStringAppend(&newList, RatAddressMail(adrPtr), -1);
-		    Tcl_DStringAppend(&newList, " (", 2);
-		    Tcl_DStringAppend(&newList, userPtr, -1);
-		    Tcl_DStringAppend(&newList, ")", 1);
-
-		} else {
-		    Tcl_DStringAppend(&newList, RatAddressMail(adrPtr), -1);
-		    if (adrPtr->personal && strlen(adrPtr->personal)) {
-			sprintf(buf, " (%.1020s)", adrPtr->personal);
-			Tcl_DStringAppend(&newList, buf, -1);
-		    }
-		}
-		if (adrPtr->next && adrPtr->next->mailbox) {
-		    Tcl_DStringAppend(&newList, ", ", 2);
-		}
-	    }
-	    mail_free_address(&baseAdrPtr);
-	} while(rescan);
-	Tcl_DStringResult(interp, &newList);
+	Tcl_DStringInit(&list);
+	Tcl_DStringAppend(&list, Tcl_GetString(objv[2]), -1);
+	RatExpandAlias(interp, &list, &ae);
+	Tcl_DStringResult(interp, &list);
 	return TCL_OK;
-	
     } else {
 	Tcl_AppendResult(interp, "bad option \"", Tcl_GetString(objv[1]),
 		"\": must be one of add, delete, get, list, read, save,",
@@ -965,59 +908,14 @@ RatAddressMail(ADDRESS *adrPtr)
 {
     static char *store = NULL;
     static int length = 0;
-    int quote = 0, grow, l, i;
-    char *cPtr;
+    size_t size = RatAddressSize(adrPtr, 1);
 
-    /*
-     * We start by checking if the mailbox part contains any special
-     * chracters (actually specials, SPACE and CTLS). While we do that we
-     * also calculate how much it will grow when converted to qtext.
-     * If we must quote it then we do so.
-     * After that we append the host part if any.
-     */
-    for (cPtr = adrPtr->mailbox, grow=0; *cPtr; cPtr++) {
-	if (strchr("\"\\\n", *cPtr)) {
-	    quote = 1;
-	    grow++;
-	}
-	if (*cPtr <= 32 || strchr("()<>@,;:[]", *cPtr)) {
-	    quote = 1;
-	}
+    if (size > length) {
+	length = size+1024;
+	store = ckrealloc(store, length);
     }
-    if (quote) {
-	l = 2+strlen(adrPtr->mailbox)+grow
-		+(adrPtr->host?strlen(adrPtr->host)+1:0)+1;
-	if (l > length) {
-	    length = l;
-	    store = ckrealloc(store, length);
-	}
-	i = 0;
-	store[i++] = '"';
-	for (cPtr = adrPtr->mailbox, grow=0; *cPtr; cPtr++) {
-	    if (strchr("\"\\\n", *cPtr)) {
-		store[i++] = '\\';
-	    }
-	    store[i++] = *cPtr;
-	}
-	store[i++] = '"';
-    } else {
-	l = strlen(adrPtr->mailbox)+(adrPtr->host?strlen(adrPtr->host)+1:0)+1;
-	if (l > length) {
-	    length = l;
-	    store = ckrealloc(store, length);
-	}
-	strcpy(store, adrPtr->mailbox);
-	i = strlen(store);
-    }
-    if (adrPtr->host && *adrPtr->host) {
-	store[i++] = '@';
-	RatStrNCpy(&store[i], adrPtr->host, length-i);
-    } else {
-	store[i++] = '\0';
-    }
-#ifdef MEM_DEBUG
-    mem_store = store;
-#endif /* MEM_DEBUG */
+    store[0] = '\0';
+    rfc822_address(store, adrPtr);
     return store;
 }
 
@@ -1039,75 +937,39 @@ RatAddressMail(ADDRESS *adrPtr)
  */
  
 int
-RatSplitAddresses(ClientData clientData, Tcl_Interp *interp, int argc,
-	char *argv[])
+RatSplitAddresses(ClientData clientData, Tcl_Interp *interp, int objc,
+		  Tcl_Obj *const objv[])
 {
-    char *srcPtr, *dstPtr, *cPtr, *adr;
-    Tcl_DString result;
-    int inq = 0, inc = 0;
+    Tcl_Obj *rPtr;
+    CONST84 char *s, *e, *n;
  
-    if (argc != 2) {
-	Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
-		" addresslist\"", (char *) NULL);
+    if (objc != 2) {
+	Tcl_AppendResult(interp, "wrong # args: should be \"",
+			 Tcl_GetString(objv[0]), " addresslist\"",
+			 (char *) NULL);
         return TCL_ERROR;
     }
- 
-    adr = (char*)ckalloc(strlen(argv[1])+1);
-    Tcl_DStringInit(&result);
- 
-    for (srcPtr = argv[1], dstPtr = adr; *srcPtr; srcPtr++) {
-	switch (*srcPtr) {
-	case '\\':
-	    if (srcPtr[1]) {
-		*dstPtr++ = *srcPtr++;
-	    }
-	    *dstPtr++ = *srcPtr;
-	    break;
-	case '"': 
-	    if (inq) {
-		inq = 0;
-	    } else {
-		inq = 1;
-	    }
-	    *dstPtr++ = *srcPtr;
-	    break;
-	case '(':
-	    inc = 1;
-	    *dstPtr++ = *srcPtr;
-	    break;
-	case ')':
-	    inc = 0;
-	    *dstPtr++ = *srcPtr;
-	    break;
-	case ',':
-	    if (!inq && !inc) {
-		for (dstPtr--;
-			dstPtr >= adr && isspace((unsigned char)*dstPtr);
-			dstPtr--);
-		if (++dstPtr != adr) {
-		    *dstPtr = '\0';
-		    for (cPtr = adr;
-			    *cPtr && isspace((unsigned char)*cPtr); cPtr++);
-		    if (cPtr) {
-			Tcl_DStringAppendElement(&result, cPtr);
-		    }
-		    dstPtr = adr;
-		}
-		break;
-	    }
-	    /* fallthrough */
-	default:
-	    *dstPtr++ = *srcPtr;
-	    break;
+
+    rPtr = Tcl_NewObj();
+    s = Tcl_GetString(objv[1]);
+    while (*s) {
+	while (*s && isspace(*s)) {
+	    s++;
+	}
+	e = n = RatFindCharInHeader(s, ',');
+	if (NULL == e) {
+	    e = n = s+strlen(s);
+	}
+	for (e--; isspace(*e) && e>s; e--) {
+	    /* Do nothing */
+	}
+	Tcl_ListObjAppendElement(interp, rPtr, Tcl_NewStringObj(s, e-s+1));
+	s = n;
+	if (*s) {
+	    s++;
 	}
     }
-    if (dstPtr != adr) {
-	*dstPtr = '\0';
-	Tcl_DStringAppendElement(&result, adr);
-    }
- 
-    Tcl_DStringResult(interp, &result);
-    ckfree(adr);
+    Tcl_SetObjResult(interp, rPtr);
     return TCL_OK;
 }
 
@@ -1117,14 +979,12 @@ void ratAddressCleanup()
     Tcl_HashEntry *e;
     Tcl_HashSearch s;
 
-    for (e = Tcl_FirstHashEntry(&aliasTable, &s); e; e = Tcl_NextHashEntry(&s)){
+    for (e = Tcl_FirstHashEntry(&aliasTable, &s); e; e =Tcl_NextHashEntry(&s)){
 	ckfree(Tcl_GetHashValue(e));
     }
     Tcl_DeleteHashTable(&aliasTable);
 
-    if (mem_store) {
-	ckfree(mem_store);
-    }
+    ckfree(mem_store);
 }
 #endif /* MEM_DEBUG */
 
@@ -1170,3 +1030,352 @@ RatAddressSize(ADDRESS *adrPtr, int all)
     return len;
 }
 
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * RatGenerateAddresses --
+ *
+ *	Generates addresses to be used when sending email
+ *
+ * Results:
+ *	The generated addresses are left in the ADDRESS-pointers
+ *	(which are assumed to not point to anything)
+ *
+ * Side effects:
+ *	None.
+ *
+ *
+ *----------------------------------------------------------------------
+ */
+void
+RatGenerateAddresses(Tcl_Interp *interp, const char *role, char *msgh,
+		     ADDRESS **from, ADDRESS **sender)
+{
+    char *tmp, host[1024];
+    const char *ctmp;
+    Tcl_Obj *oPtr;
+    int useFrom, cs;
+
+    strlcpy(host, RatGetCurrent(interp, RAT_HOST, role), sizeof(host));
+
+    *from = NULL;
+    *sender = NULL;
+
+    oPtr = Tcl_GetVar2Ex(interp, "option", "use_from", TCL_GLOBAL_ONLY);
+    if (TCL_OK != Tcl_GetBooleanFromObj(interp, oPtr, &useFrom)) {
+	useFrom = 0;
+    }
+    if (useFrom && (ctmp = Tcl_GetVar2(interp, msgh, "from",TCL_GLOBAL_ONLY))
+	    && !RatIsEmpty(ctmp)) {
+	tmp = cpystr(ctmp);
+	rfc822_parse_adrlist(from, tmp, host);
+	ckfree(tmp);
+    }
+    oPtr = Tcl_GetVar2Ex(interp, "option","create_sender",TCL_GLOBAL_ONLY);
+    Tcl_GetBooleanFromObj(interp, oPtr, &cs);
+    if (*from && cs) {
+	ADDRESS *adrPtr;
+
+	for (adrPtr = *from; adrPtr; adrPtr = adrPtr->next) {
+	    if (RatAddressIsMe(interp, adrPtr, 0)) {
+		break;
+	    }
+	}
+	if (!adrPtr) {
+	    *sender = mail_newaddr();
+	    (*sender)->personal =
+		cpystr(RatGetCurrent(interp, RAT_PERSONAL, role));
+	    (*sender)->mailbox =
+		cpystr(RatGetCurrent(interp, RAT_MAILBOX, role));
+	    (*sender)->host = cpystr(host);
+	    RatEncodeAddresses(interp, *sender);
+	}
+    } else if (!*from) {
+	*from = mail_newaddr();
+	(*from)->personal =
+	    cpystr(RatGetCurrent(interp, RAT_PERSONAL, role));
+	(*from)->mailbox =
+	    cpystr(RatGetCurrent(interp, RAT_MAILBOX, role));
+	(*from)->host = cpystr(host);
+    }
+    RatEncodeAddresses(interp, *from);
+}
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * RatGenerateAddressesCmd --
+ *
+ *      See ../doc/interface for documentation.
+ *
+ * Results:
+ *	A standard tcl result.
+ *
+ * Side effects:
+ *	None.
+ *
+ *
+ *----------------------------------------------------------------------
+ */
+
+int
+RatGenerateAddressesCmd(ClientData clientData, Tcl_Interp *interp, int objc,
+			Tcl_Obj *const objv[])
+{
+    ADDRESS *from, *sender;
+    const char *role;
+    char buf[1024];
+    Tcl_Obj *oPtr;
+    
+    if (2 != objc) {
+	Tcl_AppendResult(interp, "wrong # args: should be \"",
+		Tcl_GetString(objv[0]), " handler\"", (char *) NULL);
+	return TCL_ERROR;
+    }
+
+    role = Tcl_GetVar2(interp, Tcl_GetString(objv[1]), "role",TCL_GLOBAL_ONLY);
+    RatGenerateAddresses(interp, role, Tcl_GetString(objv[1]),
+			 &from, &sender);
+    oPtr = Tcl_NewObj();
+    buf[0] = '\0';
+    rfc822_write_address_full(buf, from, NULL);
+    Tcl_ListObjAppendElement(interp, oPtr, Tcl_NewStringObj(buf, -1));
+    buf[0] = '\0';
+    rfc822_write_address_full(buf, sender, NULL);
+    Tcl_ListObjAppendElement(interp, oPtr, Tcl_NewStringObj(buf, -1));
+    buf[0] = '\0';
+    mail_free_address(&from);
+    mail_free_address(&sender);
+    Tcl_SetObjResult(interp, oPtr);
+    return TCL_OK;
+}
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * RatExpandAlias --
+ *
+ *      Expand an alias list
+ *
+ * Results:
+ *	None
+ *
+ * Side effects:
+ *	May modify the given list.
+ *
+ *
+ *----------------------------------------------------------------------
+ */
+static void
+RatExpandAlias(Tcl_Interp *interp, Tcl_DString *list, AliasExpand *ae)
+{
+    static char *buf = NULL;
+    static int bufsize = 0;
+    CONST84 char *entry_start, *entry_end, *key_start, *key_end, *c, *copy, *n;
+    struct passwd *pwPtr;
+    int length;
+    AliasInfo *aliasPtr;
+    Tcl_HashEntry *entryPtr;
+    Tcl_DString expanded;
+    ParseState ps;
+
+    /*
+     * Here we make a copy of the list of addresses.
+     * Then we loop over all addresses and try to expand them.
+     * After we have expanded an alias we try to expand the expansion as well.
+     */
+    copy = cpystr(Tcl_DStringValue(list));
+    Tcl_DStringSetLength(list, 0);
+    for (n = entry_start = copy; *n; entry_start = n+1) {
+	/*
+	 * Find addresses in string. We assume addresses are delimited by ','.
+	 * We will also ignore any comments in addresses.
+	 */
+	while (isspace(*(unsigned char*)entry_start)) {
+	    entry_start++;
+	}
+	ps = STATE_NORMAL;
+	n = RatFindCharInHeader(entry_start, ',');
+	if (NULL == n) {
+	    n = entry_start + strlen(entry_start);
+	}
+	for (entry_end = n-1;
+	     entry_end >= entry_start && isspace((unsigned char)*entry_end);
+	     entry_end--);
+	if (entry_start > entry_end) {
+	    continue;
+	}
+	key_start = entry_start;
+	while ('(' == *key_start && key_start < entry_end) {
+	    for (key_start++; *key_start && ')' != *key_start;key_start++);
+	    for (key_start++; *key_start&&isspace(*key_start);key_start++);
+	}
+	key_end = entry_end;
+	while (')' == *key_end && key_end > key_start) {
+	    for (; key_end >= key_start && '(' != *key_end; key_end--);
+	    for (;
+		 key_end >= key_start && (isspace(*key_end) || '(' ==*key_end);
+		 key_end--);
+	}
+	if (bufsize < (key_end-key_start+1)) {
+	    bufsize = key_end-key_start+256;
+	    buf = (char*)ckrealloc(buf, bufsize);
+	}
+	memcpy(buf, key_start, key_end-key_start+1);
+	buf[key_end-key_start+1] = '\0';
+
+	if (NULL != (entryPtr = Tcl_FindHashEntry(&aliasTable, buf))
+	    && (aliasPtr = (AliasInfo*)Tcl_GetHashValue(entryPtr))
+	    && aliasPtr->mark != aliasMark) {
+	    /*
+	     * Found entry in alias database
+	     */
+	    aliasPtr->mark = aliasMark;
+	    Tcl_DStringInit(&expanded);
+	    switch (ae->level) {
+	    case 0:
+		Tcl_DStringAppend(&expanded, Tcl_GetString(aliasPtr->content),
+				  -1);
+		break;
+	    case 1:
+		Tcl_DStringAppend(&expanded, key_start,
+				  key_end-key_start+1);
+		if (Tcl_GetCharLength(aliasPtr->fullname)) {
+		    Tcl_DStringAppend(&expanded, " (", 2);
+		    Tcl_DStringAppend(&expanded,
+				      Tcl_GetString(aliasPtr->fullname), -1);
+		    Tcl_DStringAppend(&expanded, ")", 1);
+		}
+		break;
+	    case 2:
+		Tcl_DStringAppend(&expanded, Tcl_GetString(aliasPtr->content),
+				  -1);
+		if (Tcl_GetCharLength(aliasPtr->fullname)
+		    && NULL == strchr(Tcl_GetString(aliasPtr->content), ',')) {
+		    Tcl_DStringAppend(&expanded, " (", 2);
+		    Tcl_DStringAppend(&expanded,
+				      Tcl_GetString(aliasPtr->fullname), -1);
+		    Tcl_DStringAppend(&expanded, ")", 1);
+		}
+	    }
+	    RatExpandAlias(interp, &expanded, ae);
+	    Tcl_DStringAppend(list, Tcl_DStringValue(&expanded),
+			      Tcl_DStringLength(&expanded));
+	    Tcl_DStringFree(&expanded);
+	    
+	} else if (ae->lookup_in_passwd && NULL != (pwPtr = getpwnam(buf))
+	    && (NULL == entryPtr || NULL == aliasPtr)) {
+	    /*
+	     * Found user in the passwd-database
+	     */
+	    switch (ae->level) {
+	    case 0:
+		Tcl_DStringAppend(list, pwPtr->pw_name, -1);
+		break;
+	    case 1:
+		Tcl_DStringAppend(list, pwPtr->pw_name, -1);
+		Tcl_DStringAppend(list, " (", 2);
+		if ((c = strchr(pwPtr->pw_gecos, ','))) {
+		    length = c - pwPtr->pw_gecos;
+		} else {
+		    length = strlen(pwPtr->pw_gecos);
+		}
+		Tcl_DStringAppend(list, pwPtr->pw_gecos, length);
+		Tcl_DStringAppend(list, ")", 1);
+		break;
+	    case 2:
+		Tcl_DStringAppend(list, pwPtr->pw_name, -1);
+		Tcl_DStringAppend(list, "@", 1);
+		Tcl_DStringAppend(list, ae->host, -1);
+		Tcl_DStringAppend(list, " (", 2);
+		if ((c = strchr(pwPtr->pw_gecos, ','))) {
+		    length = c - pwPtr->pw_gecos;
+		} else {
+		    length = strlen(pwPtr->pw_gecos);
+		}
+		Tcl_DStringAppend(list, pwPtr->pw_gecos, length);
+		Tcl_DStringAppend(list, ")", 1);
+		break;
+	    }
+	} else {
+	    /*
+	     * User not found anywhere
+	     */
+	    Tcl_DStringAppend(list, entry_start, entry_end-entry_start+1);
+	}
+	Tcl_DStringAppend(list, ", ", 2);
+    }
+    ckfree(copy);
+    length = Tcl_DStringLength(list);
+    Tcl_DStringSetLength(list, length-2);
+}
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * RatFindCharInHeader --
+ *
+ *      Finds the next unquoted instance of a given character in a header
+ *      field.
+ *
+ * Results:
+ *	Returns the address of the next instance of the sought character
+ *      or NULL if no instance was found.
+ *
+ * Side effects:
+ *	None.
+ *
+ *
+ *----------------------------------------------------------------------
+ */
+CONST84 char*
+RatFindCharInHeader(CONST84 char *header, char m)
+{
+    ParseState ps;
+    CONST84 char *c;
+    
+    ps = STATE_NORMAL;
+    for (c = header; *c; c++) {
+	switch (ps) {
+	case STATE_NORMAL:
+	    if ('"' == *c) {
+		ps = STATE_QUOTED;
+	    } else if ('[' == *c) {
+		ps = STATE_LITERAL;
+	    } else if ('(' == *c) {
+		ps = STATE_COMMENT;
+	    } else if ('\\' == *c) {
+		ps = STATE_ESCAPED;
+	    } else if (m == *c) {
+		return c;
+	    }
+	    break;
+	case STATE_ESCAPED:
+	    ps = STATE_NORMAL;
+	    break;
+	case STATE_COMMENT:
+	    if (')' == *c) {
+		ps = STATE_NORMAL;
+	    } else if ('\\' == *c) {
+		ps = STATE_ESCAPED;
+	    }
+	    break;
+	case STATE_QUOTED:
+	    if ('"' == *c) {
+		ps = STATE_NORMAL;
+	    } else if ('\\' == *c) {
+		ps = STATE_ESCAPED;
+	    }
+	    break;
+	case STATE_LITERAL:
+	    if (']' == *c) {
+		ps = STATE_NORMAL;
+	    } else if ('\\' == *c) {
+		ps = STATE_ESCAPED;
+	    }
+	    break;
+	}
+    }
+    return NULL;
+}
