@@ -3,7 +3,7 @@
  *
  *	This file handles the delivery status notifications.
  *
- * TkRat software and its included text is Copyright 1996-2000 by
+ * TkRat software and its included text is Copyright 1996-2002 by
  * Martin Forssn
  *
  * The full text of the legal notice is contained in the file called
@@ -14,13 +14,13 @@
 #include <unistd.h>
 
 typedef struct {
-    char *envid;		/* Original envelope ID                       */
-    Tcl_DString msgFields;	/* Contains tcl list of msg fields            */
-    int numRecipients;		/* Number of recipients mentioned in this DSN */
-    char **actionPtrPtr;	/* The action                                 */
-    char **recTypePtrPtr;	/* Type of recipient addresses                */
-    char **recipientPtrPtr;	/* Recipient addresses                        */
-    Tcl_DString *rListPtrPtr;   /* Recipient fields                           */
+    char *envid;		/* Original envelope ID                      */
+    Tcl_Obj *msgFields;		/* Contains tcl list of msg fields           */
+    int numRecipients;		/* Number of recipients mentioned in this DSN*/
+    char **actionPtrPtr;	/* The action                                */
+    char **recTypePtrPtr;	/* Type of recipient addresses               */
+    char **recipientPtrPtr;	/* Recipient addresses                       */
+    Tcl_Obj **rListPtrPtr;      /* Recipient fields                          */
 } RatDeliveryStatus;
 
 /*
@@ -32,12 +32,11 @@ static Tcl_HashTable seenTable;
  * Local functions
  */
 static Tcl_Channel OpenIndex(Tcl_Interp *interp, char *mode);
-static int RatDSNList(ClientData clientData, Tcl_Interp *interp,
-	int argc, char **argv);
-static int RatDSNGet(ClientData clientData, Tcl_Interp *interp,
-	int argc, char **argv);
-static int RatDSNExpire(Tcl_Interp *interp, char *line);
-static char *RatParseDSNLine(char *buf, char *name, char *value, int *length);
+static Tcl_ObjCmdProc RatDSNList;
+static Tcl_ObjCmdProc RatDSNGet;
+static int RatDSNExpire(Tcl_Interp *interp, Tcl_Obj *lineObj);
+static char *RatParseDSNLine(char *buf, Tcl_Obj **name, Tcl_Obj **value,
+			     int *length);
 static RatDeliveryStatus *RatParseDS(Tcl_Interp *interp, Tcl_Obj *body);
 static void RatFreeDeliveryStatus(RatDeliveryStatus *statusPtr);
 
@@ -64,8 +63,8 @@ int
 RatDSNInit(Tcl_Interp *interp)
 {
     Tcl_InitHashTable(&seenTable, TCL_STRING_KEYS);
-    Tcl_CreateCommand(interp, "RatDSNList", RatDSNList, NULL, NULL);
-    Tcl_CreateCommand(interp, "RatDSNGet", RatDSNGet, NULL, NULL);
+    Tcl_CreateObjCommand(interp, "RatDSNList", RatDSNList, NULL, NULL);
+    Tcl_CreateObjCommand(interp, "RatDSNGet", RatDSNGet, NULL, NULL);
     return TCL_OK;
 }
 
@@ -87,7 +86,7 @@ RatDSNInit(Tcl_Interp *interp)
  */
 
 DSNhandle
-RatDSNStartMessage(Tcl_Interp *interp, char *id, char *subject)
+RatDSNStartMessage(Tcl_Interp *interp, const char *id, const char *subject)
 {
     Tcl_DString *dsPtr = (Tcl_DString*)ckalloc(sizeof(Tcl_DString));
     char buf[32], *header, *cPtr;
@@ -221,14 +220,15 @@ RatDSNFinish(Tcl_Interp *interp, DSNhandle handle)
 int
 RatDSNHandle (Tcl_Interp *interp, char *msg)
 {
-    int i, j, new, match = 0, changed = 0, argc, rfound, perm, rargc;
+    int i, j, new, match = 0, changed = 0, objc, rfound, perm, robjc, srobjc;
     RatDeliveryStatus *statusPtr;
     Tcl_HashEntry *entryPtr;
     Tcl_Channel channel;
-    char buf[1024], id[1024], **argv, *oldId, *dir, *msgFile = NULL;
-    char **rargv, *file;
-    Tcl_DString ds, dirDS, line;
+    char buf[1024], id[1024], *oldId, *msgFile = NULL, *file;
+    const char *dir;
     Tcl_CmdInfo cmdInfo;
+    Tcl_Obj *oPtr, *linePtr, *l1Ptr, *l2Ptr, *l3Ptr, *l4Ptr, **objv, **robjv,
+	**srobjv;
 
     /*
      * Avoid processing the same DSN twice
@@ -239,7 +239,7 @@ RatDSNHandle (Tcl_Interp *interp, char *msg)
     }
     Tcl_SetHashValue(entryPtr, 0);
 
-    sprintf(buf, "[lindex [[%s body] children] 1] data 0", msg);
+    snprintf(buf, sizeof(buf), "[lindex [[%s body] children] 1] data 0", msg);
     if (TCL_OK != Tcl_Eval(interp, buf)) {
 	return 0;
     }
@@ -249,60 +249,62 @@ RatDSNHandle (Tcl_Interp *interp, char *msg)
 	return 0;
     }
 
-    Tcl_DStringInit(&ds);
-    Tcl_DStringInit(&line);
     if (NULL == (channel = OpenIndex(interp, "r"))) {
 	RatFreeDeliveryStatus(statusPtr);
 	return 0;
     }
-    Tcl_GetInt(interp, Tcl_GetVar2(interp, "option", "permissions",
-	    TCL_GLOBAL_ONLY), &perm);
-    dir = Tcl_TranslateFileName(interp, Tcl_GetVar2(interp, "option",
-	    "dsn_directory", TCL_GLOBAL_ONLY), &dirDS);
-    while (Tcl_DStringSetLength(&line, 0), -1 != Tcl_Gets(channel, &line)) {
-	if (RatDSNExpire(interp, Tcl_DStringValue(&line))) {
+    l1Ptr = Tcl_NewObj();
+    oPtr = Tcl_GetVar2Ex(interp, "option", "permissions", TCL_GLOBAL_ONLY);
+    Tcl_GetIntFromObj(interp, oPtr, &perm);
+    dir = RatGetPathOption(interp, "dsn_directory");
+    while (linePtr = Tcl_NewObj(), -1 != Tcl_GetsObj(channel, linePtr)) {
+	if (RatDSNExpire(interp, linePtr)) {
 	    /*
 	     * This DSN has expired so we should remove all associated files
 	     */
-	    file = RatLindex(interp, Tcl_DStringValue(&line), 0);
-	    snprintf(buf, sizeof(buf), "%s/%s", dir, file);
+	    Tcl_ListObjIndex(interp, linePtr, 0, &oPtr);
+	    snprintf(buf, sizeof(buf), "%s/%s", dir, Tcl_GetString(oPtr));
 	    (void)unlink(buf);
-	    Tcl_SplitList(interp, RatLindex(interp, Tcl_DStringValue(&line), 3),
-		    &argc, &argv);
-	    for (i=0; i < argc; i++) {
-		file = RatLindex(interp, argv[i], 2);
+	    Tcl_ListObjLength(interp, linePtr, &i);
+	    Tcl_ListObjIndex(interp, linePtr, i-1, &oPtr);
+	    Tcl_ListObjGetElements(interp, oPtr, &objc, &objv);
+	    for (i=0; i < objc; i++) {
+		Tcl_ListObjIndex(interp, objv[i], 2, &oPtr);
+		file = Tcl_GetString(oPtr);;
 		if (strlen(file)) {
 		    snprintf(buf, sizeof(buf), "%s/%s", dir, file);
 		    (void)unlink(buf);
 		}
 	    }
-	    ckfree(argv);
 	    changed++;
 	    continue;
 	}
-	if (strncmp(Tcl_DStringValue(&line), statusPtr->envid,
-		strlen(statusPtr->envid))) {
-	    Tcl_DStringAppendElement(&ds, Tcl_DStringValue(&line));
+	Tcl_ListObjIndex(interp, linePtr, 0, &oPtr);
+	if (strcmp(Tcl_GetString(oPtr), statusPtr->envid)) {
+	    Tcl_ListObjAppendElement(interp, l1Ptr, linePtr);
 	    continue;
 	}
 	changed++;
 	match = 1;
-	Tcl_DStringStartSublist(&ds);
-	Tcl_SplitList(interp, Tcl_DStringValue(&line), &argc, &argv);
-	Tcl_DStringAppendElement(&ds, argv[0]);
-	Tcl_DStringAppendElement(&ds, argv[1]);
-	Tcl_DStringAppendElement(&ds, argv[2]);
-	Tcl_SplitList(interp, argv[3], &rargc, &rargv);
-	Tcl_DStringStartSublist(&ds);
-	for (i=0; i<rargc; i++) {
+	l2Ptr = Tcl_NewObj();
+	for (i=0; i<3; i++) {
+	    Tcl_ListObjIndex(interp, linePtr, i, &oPtr);
+	    Tcl_ListObjAppendElement(interp, l2Ptr, oPtr);
+	}
+	l3Ptr = Tcl_NewObj();
+	Tcl_ListObjLength(interp, linePtr, &i);
+	Tcl_ListObjIndex(interp, linePtr, i-1, &oPtr);
+	Tcl_ListObjGetElements(interp, oPtr, &robjc, &robjv);
+	for (i=0; i<robjc; i++) {
 	    for (j=rfound=0; !rfound && j<statusPtr->numRecipients; j++) {
+		Tcl_ListObjGetElements(interp, robjv[i], &srobjc, &srobjv);
 		if (statusPtr->recTypePtrPtr[j]
 			&& statusPtr->actionPtrPtr[j]
 			&& !strcasecmp(statusPtr->recTypePtrPtr[j], "rfc822")
 			&& !strcmp(statusPtr->recipientPtrPtr[j],
-			        RatLindex(interp, rargv[i], 1))
+				   Tcl_GetString(srobjv[1]))
 			&& strcmp(statusPtr->actionPtrPtr[j],
-			        RatLindex(interp, rargv[i], 0))) {
+				  Tcl_GetString(srobjv[0]))) {
 		    /*
 		     * This DSN matched this recipient
 		     * We start by saving the DSN message;
@@ -310,9 +312,9 @@ RatDSNHandle (Tcl_Interp *interp, char *msg)
 		     * Finally we notify the user.
 		     */
 		    rfound = 1;
-		    oldId = RatLindex(interp, rargv[i], 2);
+		    oldId = Tcl_GetString(srobjv[2]);
 		    RatGenId(NULL, interp, 0, NULL);
-		    RatStrNCpy(id, Tcl_GetStringResult(interp), sizeof(id));
+		    strlcpy(id, Tcl_GetStringResult(interp), sizeof(id));
 		    if (strlen(oldId)) {
 			snprintf(buf, sizeof(buf), "%s/%s", dir, oldId);
 			(void)unlink(buf);
@@ -329,7 +331,7 @@ RatDSNHandle (Tcl_Interp *interp, char *msg)
 			RatMessageGet(interp,
 				      (MessageInfo*)cmdInfo.objClientData,
 				      &msgDS, NULL, 0, NULL, 0);
-			msgCh = Tcl_OpenFileChannel(interp, msgFile, "w", perm);
+			msgCh = Tcl_OpenFileChannel(interp, msgFile, "w",perm);
 			Tcl_Write(msgCh, Tcl_DStringValue(&msgDS),
 				Tcl_DStringLength(&msgDS));
 			Tcl_Close(interp, msgCh);
@@ -337,24 +339,28 @@ RatDSNHandle (Tcl_Interp *interp, char *msg)
 		    } else {
 			link(msgFile, buf);
 		    }
-		    Tcl_DStringStartSublist(&ds);
-		    Tcl_DStringAppendElement(&ds, statusPtr->actionPtrPtr[j]);
-		    Tcl_DStringAppendElement(&ds,statusPtr->recipientPtrPtr[j]);
-		    Tcl_DStringAppendElement(&ds, id);
-		    Tcl_DStringEndSublist(&ds);
-		    Tcl_VarEval(interp, "RatDSNRecieve {", argv[2], "} {",
-			    statusPtr->actionPtrPtr[j], "} {",
-			    statusPtr->recipientPtrPtr[j], "} {", id, "}",NULL);
+		    l4Ptr = Tcl_NewObj();
+		    oPtr = Tcl_NewStringObj(statusPtr->actionPtrPtr[j], -1);
+		    Tcl_ListObjAppendElement(interp, l4Ptr, oPtr);
+		    oPtr = Tcl_NewStringObj(statusPtr->recipientPtrPtr[j], -1);
+		    Tcl_ListObjAppendElement(interp, l4Ptr, oPtr);
+		    oPtr = Tcl_NewStringObj(id, -1);
+		    Tcl_ListObjAppendElement(interp, l4Ptr, oPtr);
+		    Tcl_ListObjAppendElement(interp, l3Ptr, l4Ptr);
+		    Tcl_ListObjIndex(interp, linePtr, 2, &oPtr);
+		    Tcl_VarEval(interp, "RatDSNRecieve {",
+				Tcl_GetString(oPtr), "} {",
+				statusPtr->actionPtrPtr[j], "} {",
+				statusPtr->recipientPtrPtr[j], "} {", id,
+				"}",NULL);
 		}
 	    }
 	    if (!rfound) {
-		Tcl_DStringAppendElement(&ds, rargv[i]);
+		Tcl_ListObjAppendElement(interp, l3Ptr, robjv[i]);
 	    }
 	}
-	Tcl_DStringEndSublist(&ds);
-	Tcl_DStringEndSublist(&ds);
-	ckfree(argv);
-	ckfree(rargv);
+	Tcl_ListObjAppendElement(interp, l2Ptr, l3Ptr);
+	Tcl_ListObjAppendElement(interp, l1Ptr, l2Ptr);
     }
     Tcl_Close(interp, channel);
     RatFreeDeliveryStatus(statusPtr);
@@ -362,21 +368,17 @@ RatDSNHandle (Tcl_Interp *interp, char *msg)
 	if (NULL == (channel = OpenIndex(interp, "w"))) {
 	    return 0;
 	}
-	Tcl_SplitList(interp, Tcl_DStringValue(&ds), &argc, &argv);
-	for (i=0; i<argc; i++) {
-	    Tcl_Write(channel, argv[i], strlen(argv[i]));
+	Tcl_ListObjGetElements(interp, l1Ptr, &objc, &objv);
+	for (i=0; i<objc; i++) {
+	    Tcl_WriteObj(channel, objv[i]);
 	    Tcl_Write(channel, "\n", 1);
 	}
-	ckfree(argv);
 	Tcl_Close(interp, channel);
     }
-    Tcl_DStringFree(&dirDS);
-    Tcl_DStringFree(&ds);
-    Tcl_DStringFree(&line);
+    Tcl_DecrRefCount(l1Ptr);
+    Tcl_DecrRefCount(linePtr);
     Tcl_SetHashValue(entryPtr, match);
-    if (msgFile) {
-	ckfree(msgFile);
-    }
+    ckfree(msgFile);
     return match;
 }
 
@@ -401,17 +403,17 @@ int
 RatDSNExtract (Tcl_Interp *interp, Tcl_Obj *body)
 {
     RatDeliveryStatus *sPtr = RatParseDS(interp, body);
-    Tcl_DString ds;
+    Tcl_Obj *rPtr, *oPtr;
     int i;
 
-    Tcl_DStringInit(&ds);
-    Tcl_DStringAppendElement(&ds, Tcl_DStringValue(&sPtr->msgFields));
-    Tcl_DStringStartSublist(&ds);
+    rPtr = Tcl_NewObj();
+    Tcl_ListObjAppendElement(interp, rPtr, sPtr->msgFields);
+    oPtr = Tcl_NewObj();
     for (i=0; i<sPtr->numRecipients; i++) {
-	Tcl_DStringAppendElement(&ds, Tcl_DStringValue(&sPtr->rListPtrPtr[i]));
+	Tcl_ListObjAppendElement(interp, oPtr, sPtr->rListPtrPtr[i]);
     }
-    Tcl_DStringEndSublist(&ds);
-    Tcl_DStringResult(interp, &ds);
+    Tcl_ListObjAppendElement(interp, rPtr, oPtr);
+    Tcl_SetObjResult(interp, rPtr);
     RatFreeDeliveryStatus(sPtr);
     return TCL_OK;
 }
@@ -437,16 +439,16 @@ RatDSNExtract (Tcl_Interp *interp, Tcl_Obj *body)
 static Tcl_Channel
 OpenIndex(Tcl_Interp *interp, char *mode)
 {
-    Tcl_DString ds;
-    char *dir, buf[1024];
+    char buf[1024];
+    const char *dir;
     struct stat sbuf;
     Tcl_Channel channel;
     int perm;
+    Tcl_Obj *oPtr;
 
-    Tcl_GetInt(interp, Tcl_GetVar2(interp, "option", "permissions",
-	    TCL_GLOBAL_ONLY), &perm);
-    dir = Tcl_TranslateFileName(interp, Tcl_GetVar2(interp, "option",
-	    "dsn_directory", TCL_GLOBAL_ONLY), &ds);
+    oPtr = Tcl_GetVar2Ex(interp, "option", "permissions", TCL_GLOBAL_ONLY);
+    Tcl_GetIntFromObj(interp, oPtr, &perm);
+    dir = RatGetPathOption(interp, "dsn_directory");
     if (stat(dir, &sbuf)) {
 	if (mkdir(dir, perm|0100)) {
 	    Tcl_AppendResult(interp, "Failed to create directory \"",
@@ -458,7 +460,6 @@ OpenIndex(Tcl_Interp *interp, char *mode)
 	return NULL;
     }
     snprintf(buf, sizeof(buf), "%s/index", dir);
-    Tcl_DStringFree(&ds);
 
     if (NULL == (channel = Tcl_OpenFileChannel(interp, buf, mode, perm))) {
 	return NULL;
@@ -485,30 +486,33 @@ OpenIndex(Tcl_Interp *interp, char *mode)
  */
 
 static int
-RatDSNList (ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
+RatDSNList (ClientData clientData, Tcl_Interp *interp, int objc,
+	    Tcl_Obj *const objv[])
 {
     Tcl_Channel channel = OpenIndex(interp, "r");
-    Tcl_DString line;
+    Tcl_Obj *oPtr, *rPtr;
 
     if (!channel) {
 	Tcl_ResetResult(interp);
 	return TCL_OK;
     }
 
-    Tcl_DStringInit(&line);
-
-    while (-1 != Tcl_Gets(channel, &line)) {
-	if (!RatDSNExpire(interp, Tcl_DStringValue(&line))) {
-	    Tcl_AppendElement(interp, Tcl_DStringValue(&line));
+    rPtr = Tcl_NewObj();
+    while (oPtr = Tcl_NewObj(), -1 != Tcl_GetsObj(channel, oPtr)) {
+	if (!RatDSNExpire(interp, oPtr)) {
+	    Tcl_ListObjAppendElement(interp, rPtr, oPtr);
+	} else {
+	    Tcl_DecrRefCount(oPtr);
 	}
-	Tcl_DStringSetLength(&line, 0);
     }
-    Tcl_DStringFree(&line);
+    Tcl_DecrRefCount(oPtr);
     if (!Tcl_Eof(channel)) {
 	Tcl_Close(interp, channel);
+	Tcl_DecrRefCount(rPtr);
 	return TCL_ERROR;
     }
     Tcl_Close(interp, channel);
+    Tcl_SetObjResult(interp, rPtr);
     return TCL_OK;
 }
 
@@ -530,37 +534,38 @@ RatDSNList (ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
  */
 
 static int
-RatDSNGet (ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
+RatDSNGet (ClientData clientData, Tcl_Interp *interp, int objc,
+	   Tcl_Obj *const objv[])
 {
-    char buf[1024], *dir, *msg, *data;
+    char buf[1024], *msg, *data;
+    const char *dir;
     RatDeliveryStatus *statusPtr;
     Tcl_Channel channel;
-    Tcl_DString ds;
     int i, len;
+    Tcl_Obj *rPtr, *oPtr;
 
-    if (argc != 3 && argc != 4) {
-	Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
-		" what id ?recipient?\"", (char *) NULL);
+    if (objc != 3 && objc != 4) {
+	Tcl_AppendResult(interp, "wrong # args: should be \"",
+			 Tcl_GetString(objv[0]), " what id ?recipient?\"",
+			 (char *) NULL);
 	return TCL_ERROR;
     }
-    if (strcmp(argv[1], "msg") && strcmp(argv[1], "report")) {
+    if (strcmp(Tcl_GetString(objv[1]), "msg")
+	&& strcmp(Tcl_GetString(objv[1]), "report")) {
 	Tcl_AppendResult(interp, "Illegal 'what' argument; should be ",
 		"'msg' or 'report'.", (char*) NULL);
 	return TCL_ERROR;
     }
-    if (!strlen(argv[2])) {
+    if (!strlen(Tcl_GetString(objv[2]))) {
 	Tcl_SetResult(interp, "Empty 'id' argument.", TCL_STATIC);
 	return TCL_ERROR;
     }
 
-    Tcl_DStringInit(&ds);
-    dir = Tcl_TranslateFileName(interp, Tcl_GetVar2(interp, "option",
-	    "dsn_directory", TCL_GLOBAL_ONLY), &ds);
-    snprintf(buf, sizeof(buf), "%s/%s", dir, argv[2]);
+    dir = RatGetPathOption(interp, "dsn_directory");
+    snprintf(buf, sizeof(buf), "%s/%s", dir, Tcl_GetString(objv[2]));
     if (NULL == (channel = Tcl_OpenFileChannel(interp, buf, "r", 0))) {
         return TCL_ERROR;
     }
-    Tcl_DStringFree(&ds);
     len = Tcl_Seek(channel, 0, SEEK_END);
     data = (char*) ckalloc(len+1);
     Tcl_Seek(channel, 0, SEEK_SET);
@@ -570,23 +575,27 @@ RatDSNGet (ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
     msg = RatFrMessageCreate(interp, data, len, NULL);
     ckfree(data);
 
-    if (!strcmp(argv[1], "msg")) {
+    if (!strcmp(Tcl_GetString(objv[1]), "msg")) {
 	Tcl_SetResult(interp, msg, TCL_VOLATILE);
     } else {
-	snprintf(buf, sizeof(buf),"[lindex [[%s body] children] 1] data 0",msg);
+	snprintf(buf,sizeof(buf),"[lindex [[%s body] children] 1] data 0",msg);
 	if (TCL_OK != Tcl_Eval(interp, buf)) {
 	    return TCL_ERROR;
 	}
 	statusPtr = RatParseDS(interp, Tcl_GetObjResult(interp));
-	Tcl_DStringResult(interp, &statusPtr->msgFields);
+	rPtr = Tcl_NewObj();
+	Tcl_ListObjAppendElement(interp, rPtr, statusPtr->msgFields);
+	oPtr = Tcl_NewObj();
 	for (i=0; i < statusPtr->numRecipients; i++) {
-	    if (!strcmp(statusPtr->recipientPtrPtr[i], argv[3])) {
-		Tcl_AppendResult(interp, " ",
-			Tcl_DStringValue(&statusPtr->rListPtrPtr[i]),
-			(char *) NULL);
+	    if (!strcmp(statusPtr->recipientPtrPtr[i],
+			Tcl_GetString(objv[3]))){
+		Tcl_ListObjAppendElement(interp, oPtr,
+			Tcl_NewStringObj(statusPtr->recipientPtrPtr[i], -1));
 		break;
 	    }
 	}
+	Tcl_ListObjAppendElement(interp, rPtr, oPtr);
+	Tcl_SetObjResult(interp, rPtr);
 	RatFreeDeliveryStatus(statusPtr);
 	RatMessageDelete(interp, msg);
     }
@@ -611,14 +620,16 @@ RatDSNGet (ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
  */
 
 static int
-RatDSNExpire(Tcl_Interp *interp, char *line)
+RatDSNExpire(Tcl_Interp *interp, Tcl_Obj *linePtr)
 {
+    Tcl_Obj *oPtr;
     long intime;
     int days;
 
-    Tcl_GetInt(interp, Tcl_GetVar2(interp, "option", "dsn_expiration",
-	    TCL_GLOBAL_ONLY), &days);
-    sscanf(line, "%*s %ld", &intime);
+    oPtr = Tcl_GetVar2Ex(interp, "option", "dsn_expiration", TCL_GLOBAL_ONLY);
+    Tcl_GetIntFromObj(interp, oPtr, &days);
+    Tcl_ListObjIndex(interp, linePtr, 1, &oPtr);
+    Tcl_GetLongFromObj(interp, oPtr, &intime);
     return (intime+days*24*60*60 < time(NULL));
 }
 
@@ -628,13 +639,10 @@ RatDSNExpire(Tcl_Interp *interp, char *line)
  * RatParseDSNLine --
  *
  *      Extract the next line of DSN information from a message/delivery-status
- *	message. The arguments name and place are expected to point to areas
- *	large enough to hold the results.
+ *	message.
  *
  * Results:
- *	The name,value pair is written into the buffers given as arguments.
- *	The function returns a pointer to the chararacter after the last one
- *	parsed.
+ *	The name and value pointers objects are filled in.
  *
  * Side effects:
  *	None.
@@ -644,49 +652,49 @@ RatDSNExpire(Tcl_Interp *interp, char *line)
  */
 
 static char*
-RatParseDSNLine(char *buf, char *name, char *value, int *length)
+RatParseDSNLine(char *line, Tcl_Obj **name, Tcl_Obj **value, int *length)
 {
-    int i, inName, preamble;
-    char *dstPtr;
+    char *s, *e, *n;
 
-    *name = *value = '\0';
-    for (i=0, dstPtr = name, inName = preamble = 1; buf[i]; i++) {
-	switch(buf[i]) {
-	case '\015':
-	    if ('\n' == buf[i+1]) {
-		break;
-	    }
-	    /* fallthrough */
-	case '\012':
-	    if (' ' != buf[i+1] && '\t' != buf[i+1]) {
-		*dstPtr = '\0';
-		*length -= i+1;
-		return buf+i+1;
-	    }
-	    break;
-	case ':':
-	    if (inName) {
-		*dstPtr = '\0';
-		dstPtr = value;
-		inName = 0;
-		preamble = 1;
-	    } else {
-		*dstPtr++ = buf[i];
-	    }
-	    break;
-	case ' ':	/* fallthrough */
-	case '\t':
-	    if (preamble) {
-		break;
-	    }
-	default:
-	    *dstPtr++ = buf[i];
-	    preamble = 0;
-	    break;
-	}
+    *name = *value = NULL;
+
+    /* Find start of name */
+    for (s=line; ' ' == *s || '\t' == *s; s++);
+    /* Find end of name */
+    for (e=s; *e && ':' != *e && '\r' != *e && '\n' != *e; e++);
+    if (':' != *e) {
+	goto bad;
+    }
+    n = e+1;
+    for (e--; isspace(e[-1]) && e>s; e--);
+    *name = Tcl_NewStringObj(s, e-s+1);
+
+    /* Find start of value */
+    for (s=n; ' ' == *s || '\t' == *s; s++);
+    /* Find end of value */
+    for (e=s; (e=strstr(e, "\r\n")) && (' ' == e[2] || '\t' == e[2]); e++);
+    if (NULL == e) {
+	n = e = s+strlen(s);
+    } else {
+	n = e+2;
+    }
+    *value = Tcl_NewStringObj(s, e-s);
+    *length -= n-line;
+    return n;
+
+ bad:
+    if (*name) {
+	Tcl_DecrRefCount(*name);
+	*name = NULL;
+    }
+    for (e=line; (e=strstr(e, "\r\n")) && (' '==e[2] || '\t'==e[2]); e++);
+    if (e) {
+	*length -= e-line;
+	return e;
+    } else {
+	*length -= strlen(line);
+	return NULL;
     }
-    *length -= i;
-    return buf+i;
 }
 
 /*
@@ -712,12 +720,11 @@ static RatDeliveryStatus*
 RatParseDS(Tcl_Interp *interp, Tcl_Obj *bPtr)
 {
     RatDeliveryStatus *sPtr = (RatDeliveryStatus*)ckalloc(sizeof(*sPtr));
-    char *name, *value, *cPtr, *body;
+    char *cPtr, *body;
     int allocated, i, length;
+    Tcl_Obj *ov[2];
 
     body = Tcl_GetStringFromObj(bPtr, &length);
-    name = (char*)ckalloc(length+1);
-    value = (char*)ckalloc(length+1);
 
     /*
      * Parse the per message fields.
@@ -727,21 +734,22 @@ RatParseDS(Tcl_Interp *interp, Tcl_Obj *bPtr)
        body++;
        length--;
     }
-    Tcl_DStringInit(&sPtr->msgFields);
-    while (length > 0) {
-	body = RatParseDSNLine(body, name, value, &length);
-	if (!*name) {
+    sPtr->msgFields = Tcl_NewObj();
+    Tcl_IncrRefCount(sPtr->msgFields);
+    while (length > 0 && body) {
+	body = RatParseDSNLine(body, &ov[0], &ov[1], &length);
+	if (!ov[0]) {
 	    break;
 	}
-	Tcl_DStringStartSublist(&sPtr->msgFields);
-	Tcl_DStringAppendElement(&sPtr->msgFields, name);
-	Tcl_DStringAppendElement(&sPtr->msgFields, value);
-	Tcl_DStringEndSublist(&sPtr->msgFields);
-	if (!strcasecmp("original-envelope-id", name)) {
-	    sPtr->envid = cpystr(value);
+	Tcl_ListObjAppendElement(interp, sPtr->msgFields,
+				 Tcl_NewListObj(2, ov));
+	if (!strcasecmp("original-envelope-id", Tcl_GetString(ov[0]))) {
+	    sPtr->envid = cpystr(Tcl_GetString(ov[1]));
 	}
 	/* Added by Lou Ruppert to account for odd DSN implementations */
-	if (!strcasecmp("arrival-date",name) && '\n'!=*body && '\r'!=*body) {
+	if (!strcasecmp("arrival-date", Tcl_GetString(ov[0]))
+	    && '\n' != *body
+	    && '\r' != *body) {
 	    break;
 	}
     }
@@ -755,7 +763,14 @@ RatParseDS(Tcl_Interp *interp, Tcl_Obj *bPtr)
     sPtr->recipientPtrPtr = NULL;
     sPtr->rListPtrPtr = NULL;
     allocated = 0;
-    while (length > 0) {
+    while (length > 0 && body) {
+	while (isspace(*body) && length > 0) {
+	    body++;
+	    length--;
+	}
+	if (!*body) {
+	    break;
+	}
 	if (allocated <= sPtr->numRecipients) {
 	    allocated += 32;
 	    sPtr->actionPtrPtr = (char**)ckrealloc(sPtr->actionPtrPtr, 
@@ -764,31 +779,31 @@ RatParseDS(Tcl_Interp *interp, Tcl_Obj *bPtr)
 		    allocated*sizeof(char*));
 	    sPtr->recipientPtrPtr = (char**)ckrealloc(sPtr->recipientPtrPtr, 
 		    allocated*sizeof(char*));
-	    sPtr->rListPtrPtr = (Tcl_DString*)ckrealloc(sPtr->rListPtrPtr, 
+	    sPtr->rListPtrPtr = (Tcl_Obj**)ckrealloc(sPtr->rListPtrPtr, 
 		    allocated*sizeof(Tcl_DString));
 	}
 	i = sPtr->numRecipients++;
 	sPtr->actionPtrPtr[i] = NULL;
 	sPtr->recTypePtrPtr[i] = NULL;
 	sPtr->recipientPtrPtr[i] = NULL;
-	Tcl_DStringInit(&sPtr->rListPtrPtr[i]);
+	sPtr->rListPtrPtr[i] = Tcl_NewObj();
+	Tcl_IncrRefCount(sPtr->rListPtrPtr[i]);
 	while (length > 0) {
-	    body = RatParseDSNLine(body, name, value, &length);
-	    if (!*name) {
+	    body = RatParseDSNLine(body, &ov[0], &ov[1], &length);
+	    if (!ov[0]) {
 		break;
 	    }
-	    Tcl_DStringStartSublist(&sPtr->rListPtrPtr[i]);
-	    Tcl_DStringAppendElement(&sPtr->rListPtrPtr[i], name);
-	    Tcl_DStringAppendElement(&sPtr->rListPtrPtr[i], value);
-	    Tcl_DStringEndSublist(&sPtr->rListPtrPtr[i]);
-	    if (!strcasecmp("original-recipient", name)) {
-		sPtr->recTypePtrPtr[i] = cpystr(value);
-		cPtr = strchr(sPtr->recTypePtrPtr[i], ';');
-		*cPtr++ = '\0';
+	    Tcl_ListObjAppendElement(interp, sPtr->rListPtrPtr[i],
+				     Tcl_NewListObj(2, ov));
+	    if (!strcasecmp("original-recipient", Tcl_GetString(ov[0]))) {
+		sPtr->recTypePtrPtr[i] = cpystr(Tcl_GetString(ov[1]));
+		if (NULL != (cPtr = strchr(sPtr->recTypePtrPtr[i], ';'))) {
+		    *cPtr++ = '\0';
+		}
 		sPtr->recipientPtrPtr[i] = cPtr;
 	    }
-	    if (!strcasecmp("action", name)) {
-		sPtr->actionPtrPtr[i] = cpystr(value);
+	    if (!strcasecmp("action", Tcl_GetString(ov[0]))) {
+		sPtr->actionPtrPtr[i] = cpystr(Tcl_GetString(ov[1]));
 	    }
 	}
 	if (!sPtr->actionPtrPtr[i]) {
@@ -797,8 +812,6 @@ RatParseDS(Tcl_Interp *interp, Tcl_Obj *bPtr)
 	    sPtr->recipientPtrPtr[i] = NULL;
 	}
     }
-    ckfree(name);
-    ckfree(value);
 
     return sPtr;
 }
@@ -827,19 +840,13 @@ RatFreeDeliveryStatus(RatDeliveryStatus *statusPtr)
 {
     int i;
 
-    if (statusPtr->envid) {
-	ckfree(statusPtr->envid);
-    }
-    Tcl_DStringFree(&statusPtr->msgFields);
+    ckfree(statusPtr->envid);
+    Tcl_DecrRefCount(statusPtr->msgFields);
     if (statusPtr->numRecipients) {
 	for (i=0; i<statusPtr->numRecipients; i++) {
-	    if (statusPtr->actionPtrPtr[i]) {
-		ckfree(statusPtr->actionPtrPtr[i]);
-	    }
-	    if (statusPtr->recTypePtrPtr[i]) {
-		ckfree(statusPtr->recTypePtrPtr[i]);
-	    }
-	    Tcl_DStringFree(&statusPtr->rListPtrPtr[i]);
+	    ckfree(statusPtr->actionPtrPtr[i]);
+	    ckfree(statusPtr->recTypePtrPtr[i]);
+	    Tcl_DecrRefCount(statusPtr->rListPtrPtr[i]);
 	}
 	ckfree(statusPtr->actionPtrPtr);
 	ckfree(statusPtr->recTypePtrPtr);
