@@ -1,7 +1,7 @@
 /*
  * ratDbase.c --
  *
- * TkRat software and its included text is Copyright 1996-2000 by
+ * TkRat software and its included text is Copyright 1996-2002 by
  * Martin Forssn
  *
  * The full text of the legal notice is contained in the file called
@@ -73,7 +73,7 @@
 
 static int isRead = 0;		/* 0 means that the database hasn't been
 				 * read yet */
-static int numRead;		/* The number of entries in the entryPtr list */
+static int numRead;		/* The number of entries in the entryPtr list*/
 static int numAlloc;		/* The number of entries that will fit
 				 * into the entryPtr list */
 static RatDbEntry *entryPtr;	/* The list of entries in the database */
@@ -107,7 +107,7 @@ static void	Update(ClientData clientData);
 static int	IsRlocked(char *ignore);
 static void	RatDbBuildList(Tcl_Interp *interp, Tcl_DString *dsPtr,
 	char *prefix, char *dir, Tcl_HashTable *tablePtr, int fix);
-static int	NoLFPrint(FILE *fp, char *s);
+static int	NoLFPrint(FILE *fp, const char *s);
 static void	DbaseConvert3to4(Tcl_Interp *interp);
 
 
@@ -152,16 +152,11 @@ Read(Tcl_Interp *interp)
      * identifier we should use.
      */
     if (0 == dbDir) {
-	Tcl_DString ds;	/* Scratch area */
-	char *value;	/* Value of option */
-
-	value = Tcl_GetVar2(interp, "option", "dbase_dir", TCL_GLOBAL_ONLY);
+	const char *value = RatGetPathOption(interp, "dbase_dir");
 	if (NULL == value) {
-	    Tcl_SetResult(interp, "option(dbase_dir) has no value", TCL_STATIC);
 	    return TCL_ERROR;
 	}
-	dbDir = cpystr(Tcl_TildeSubst(interp, value, &ds));
-	Tcl_DStringFree(&ds);
+	dbDir = cpystr(value);
     }
     if (0 == ident) {
 	gethostname(buf, sizeof(buf));
@@ -232,7 +227,7 @@ Read(Tcl_Interp *interp)
 		"\": ", Tcl_PosixError(interp), (char *) NULL);
 	return TCL_ERROR;
     }
-    (void)Tcl_CreateTimerHandler(UPDATE_INTERVAL*1000, Update,(ClientData)NULL);
+    (void)Tcl_CreateTimerHandler(UPDATE_INTERVAL*1000,Update,(ClientData)NULL);
 
     /*
      * Read the index.info file
@@ -255,8 +250,9 @@ Read(Tcl_Interp *interp)
      * complain!
      */
     if (version != DBASE_VERSION && version != 3) {
-	sprintf(buf, "wrong version of database got %d expected %d", version,
-		DBASE_VERSION);
+	snprintf(buf, sizeof(buf),
+		 "wrong version of database got %d expected %d", version,
+		 DBASE_VERSION);
 	Tcl_SetResult(interp, buf, TCL_VOLATILE);
 	goto error;
     }
@@ -315,8 +311,9 @@ Read(Tcl_Interp *interp)
 	     */
 	    if ('+' == entryPtr[i].content[EX_TYPE][0]) {
 		char buf[64];
-		sprintf(buf, "%ld", atoi(entryPtr[i].content[EX_TYPE])*24*60*60+
-			    atol(entryPtr[i].content[DATE]));
+		sprintf(buf, "%ld",
+			atoi(entryPtr[i].content[EX_TYPE])*24*60*60+
+			atol(entryPtr[i].content[DATE]));
 		entryPtr[i].content[EX_TIME] = cpystr(buf);
 		entryPtr[i].content[EX_TYPE] = "backup";
 	    }
@@ -732,7 +729,7 @@ Update(ClientData clientData)
     snprintf(rlockName, sizeof(rlockName), "%s/rlock.%s", dbDir, ident);
     (void)utime(rlockName, (struct utimbuf*) NULL);
 
-    (void)Tcl_CreateTimerHandler(UPDATE_INTERVAL*1000, Update,(ClientData)NULL);
+    (void)Tcl_CreateTimerHandler(UPDATE_INTERVAL*1000,Update,(ClientData)NULL);
 }
 
 /*
@@ -767,7 +764,7 @@ IsRlocked(char *ignore)
     while (0 != (direntPtr = readdir(dirPtr))) {
 	if (!strncmp("rlock.", (char*)direntPtr->d_name, 6)
 		&& (ignore && strcmp((char*)direntPtr->d_name, ignore))) {
-	    snprintf(buf, sizeof(buf),"%s/%s", dbDir, (char*)direntPtr->d_name);
+	    snprintf(buf, sizeof(buf),"%s/%s", dbDir,(char*)direntPtr->d_name);
 	    (void)stat(buf, &sbuf);
 	    if (deadline < sbuf.st_mtime) {
 		result = 1;
@@ -808,9 +805,11 @@ IsRlocked(char *ignore)
  */
 
 int
-RatDbInsert(Tcl_Interp *interp, char *to, char *from, char *cc, char *msgid,
-	    char *ref, char *subject, long date, char *flags, char *keywords,
-	    long exDate, char *exType, char *fromline, char *mail, int length)
+RatDbInsert (Tcl_Interp *interp, const char *to, const char *from,
+	     const char *cc, const char *msgid, const char *ref,
+	     const char *subject, long date, const char *flags,
+	     const char *keywords, long exDate, const char *exType,
+	     const char *fromline, const char *mail, int length)
 {
     static char *tobuf = NULL;	/* Scratch area */
     static int tobufsize = 0;	/* Size of scratch area */
@@ -827,12 +826,13 @@ RatDbInsert(Tcl_Interp *interp, char *to, char *from, char *cc, char *msgid,
     ADDRESS *adrPtr;		/* Address list */
     int mode;			/* Mode to use when creating files */
     int i;
+    Tcl_Obj *oPtr;
 
     /*
      * Get file creation mode
      */
-    Tcl_GetInt(interp, Tcl_GetVar2(interp, "option", "permissions",
-	    TCL_GLOBAL_ONLY), &mode);
+    oPtr = Tcl_GetVar2Ex(interp, "option", "permissions", TCL_GLOBAL_ONLY);
+    Tcl_GetIntFromObj(interp, oPtr, &mode);
 
     if (0 == isRead) {
 	if (TCL_OK != Read(interp)) {
@@ -851,14 +851,14 @@ RatDbInsert(Tcl_Interp *interp, char *to, char *from, char *cc, char *msgid,
 	    tobufsize = strlen(to)+1;
 	    tobuf = (char*)ckrealloc(tobuf, tobufsize);
 	}
-	strcpy(tobuf, to);
-	rfc822_parse_adrlist(&adrPtr, tobuf, currentHost);
+	strlcpy(tobuf, to, tobufsize);
+	rfc822_parse_adrlist(&adrPtr, tobuf, "not.used");
 	if (adrPtr && adrPtr->mailbox && *adrPtr->mailbox) {
 	    dir = cpystr(adrPtr->mailbox);
 	}
     }
     if (!dir) {
-	dir = cpystr(currentMailboxName);
+	dir = cpystr("default");
     }
     mail_free_address(&adrPtr);
     for (cPtr = dir; *cPtr; cPtr++) {
@@ -944,7 +944,7 @@ RatDbInsert(Tcl_Interp *interp, char *to, char *from, char *cc, char *msgid,
     NoLFPrint(indexFP, ref);
     NoLFPrint(indexFP, subject);
     fprintf(indexFP, "%ld\n", date);
-    NoLFPrint(indexFP, keywords);
+    NoLFPrint(indexFP, (keywords ? keywords : ""));
     fprintf(indexFP, "%d\n", length);
     NoLFPrint(indexFP, flags);
     fprintf(indexFP, "%ld\n", exDate*24*60*60 + time((time_t*) NULL));
@@ -967,8 +967,9 @@ RatDbInsert(Tcl_Interp *interp, char *to, char *from, char *cc, char *msgid,
 		"\": ", Tcl_PosixError(interp), (char *) NULL);
 	goto losing;
     }
-    Tcl_Write(dataChannel, fromline, strlen(fromline));
-    RatTranslateWrite(dataChannel, mail, length);
+    /* The casts here are needed to build on tcl <8.4 */
+    Tcl_Write(dataChannel, (char*)fromline, strlen(fromline));
+    RatTranslateWrite(dataChannel, (char*)mail, length);
     if (TCL_OK != Tcl_Close(interp, dataChannel)) {
 	Tcl_AppendResult(interp, "error closing file \"", buf,
 		"\": ", Tcl_PosixError(interp), (char *) NULL);
@@ -1108,27 +1109,30 @@ RatDbSetStatus(Tcl_Interp *interp, int index, char *status)
  */
 
 int
-RatDbSearch(Tcl_Interp *interp, char *exp, int *numFoundPtr, int **foundPtrPtr)
+RatDbSearch(Tcl_Interp *interp, Tcl_Obj *exp, int *numFoundPtr,
+	    int **foundPtrPtr)
 {
     int or;			/* Indicates the operation; 0 = and  1 = or */
     int i, j, k;		/* Loop counters */
     int match;
     int numAlloc = 0;		/* Number of entries allocated room for */
-    int expNumWords, argc;
-    char **expWords, **argv;
+    int expNumWords, objc;
+    Tcl_Obj **expWords, **objv;
     int numExp;			/* The number of subexpressions in the exp */
     char fname[1024];		/* Filename of actual message */
     int bodyfd;			/* File descriptor to actual message */
     char *message = NULL;	/* Actual message */
     int messageSize = 0;	/* Size of message area */
     struct stat sbuf;		/* Buffer for stat calls */
+    char *s;
+    
     /*
      * The following three lists describes the search expression. Every
      * expression has one entry in each list.
      */
     int *notPtr;
     RatDbEType *fieldPtr;
-    char **valuePtr;
+    Tcl_Obj **valuePtr;
 
     *numFoundPtr = 0;
     *foundPtrPtr = NULL;
@@ -1136,20 +1140,21 @@ RatDbSearch(Tcl_Interp *interp, char *exp, int *numFoundPtr, int **foundPtrPtr)
     /*
      * Parse the expression and build the lists.
      */
-    if (TCL_OK != Tcl_SplitList(interp, exp, &expNumWords, &expWords)) {
+    if (TCL_OK != Tcl_ListObjGetElements(interp, exp,&expNumWords,&expWords)) {
 	return TCL_ERROR;
     }
-    if (!strcmp(*expWords, "and") && !strcmp(*expWords, "or")) {
+    s = Tcl_GetString(expWords[0]);
+    if (!strcmp(s, "and") && !strcmp(s, "or")) {
 	Tcl_SetResult(interp, "exp must start with \"and\" or \"or\".",
 		TCL_STATIC);
 	return TCL_ERROR;
     }
     notPtr = (int*) ckalloc(sizeof(int) * (expNumWords/2));
     fieldPtr = (RatDbEType*) ckalloc(sizeof(RatDbEType) * (expNumWords/2));
-    valuePtr = (char**) ckalloc(sizeof(char*) * (expNumWords/2));
+    valuePtr = (Tcl_Obj**) ckalloc(sizeof(Tcl_Obj*) * (expNumWords/2));
 
     expNumWords--;
-    if (!strcmp(*expWords, "or")) {
+    if (!strcmp(s, "or")) {
 	or = 1;
     } else {
 	or = 0;
@@ -1158,9 +1163,11 @@ RatDbSearch(Tcl_Interp *interp, char *exp, int *numFoundPtr, int **foundPtrPtr)
     numExp = 0;
     i = 1;
     while (i < expNumWords) {
-	if (!strcmp(expWords[i], "not")) {
+	s = Tcl_GetString(expWords[i]);
+	if (!strcmp(s, "not")) {
 	    notPtr[numExp] = 1;
 	    i++;
+	    s = Tcl_GetString(expWords[i]);
 	} else {
 	    notPtr[numExp] = 0;
 	}
@@ -1170,17 +1177,17 @@ RatDbSearch(Tcl_Interp *interp, char *exp, int *numFoundPtr, int **foundPtrPtr)
 		    TCL_STATIC);
 	    goto losing;
 	}
-	if (!strcmp(expWords[i], "to")) {
+	if (!strcmp(s, "to")) {
 	    fieldPtr[numExp] = TO;
-	} else if (!strcmp(expWords[i], "from")) {
+	} else if (!strcmp(s, "from")) {
 	    fieldPtr[numExp] = FROM;
-	} else if (!strcmp(expWords[i], "cc")) {
+	} else if (!strcmp(s, "cc")) {
 	    fieldPtr[numExp] = CC;
-	} else if (!strcmp(expWords[i], "subject")) {
+	} else if (!strcmp(s, "subject")) {
 	    fieldPtr[numExp] = SUBJECT;
-	} else if (!strcmp(expWords[i], "keywords")) {
+	} else if (!strcmp(s, "keywords")) {
 	    fieldPtr[numExp] = KEYWORDS;
-	} else if (!strcmp(expWords[i], "all")) {
+	} else if (!strcmp(s, "all")) {
 	    fieldPtr[numExp] = -1;
 	} else {
 	    Tcl_SetResult(interp, "Parse error in exp (illegal field value)",
@@ -1210,8 +1217,8 @@ RatDbSearch(Tcl_Interp *interp, char *exp, int *numFoundPtr, int **foundPtrPtr)
 	}
 	match = 0;
 	for(j=0; j < numExp && !(j != 0 && or == match); j++) {
-    	    Tcl_SplitList(interp, valuePtr[j], &argc, &argv);
-	    for (k=0; k<argc && !(k != 0 && or == match); k++) {
+	    Tcl_ListObjGetElements(interp, valuePtr[j], &objc, &objv);
+	    for (k=0; k<objc && !(k != 0 && or == match); k++) {
 		if (fieldPtr[j] == -1) {
 		    snprintf(fname, sizeof(fname), "%s/dbase/%s", dbDir,
 			    entryPtr[i].content[FILENAME]);
@@ -1222,7 +1229,7 @@ RatDbSearch(Tcl_Interp *interp, char *exp, int *numFoundPtr, int **foundPtrPtr)
 			goto losing;
 		    }
 		    if (0 != fstat(bodyfd, &sbuf)) {
-			Tcl_AppendResult(interp, "error stating file \"", fname,
+			Tcl_AppendResult(interp, "error stating file \"",fname,
 				"\": ", Tcl_PosixError(interp), (char *) NULL);
 			close(bodyfd);
 			goto losing;
@@ -1234,27 +1241,27 @@ RatDbSearch(Tcl_Interp *interp, char *exp, int *numFoundPtr, int **foundPtrPtr)
 		    read(bodyfd, message, sbuf.st_size);
 		    message[sbuf.st_size] = '\0';
 		    (void)close(bodyfd);
-		    match = RatSearch(argv[k], message);
+		    match = RatSearch(Tcl_GetString(objv[k]), message);
 		} else {
-		    match = RatSearch(argv[k],entryPtr[i].content[fieldPtr[j]]);
+		    match = RatSearch(Tcl_GetString(objv[k]),
+				      entryPtr[i].content[fieldPtr[j]]);
 		}
 		if (1 == notPtr[j]) {
 		    match = match ? 0 : 1;
 		}
 	    }
-	    ckfree(argv);
 	}
 
 	if (match) {
 	    if (*numFoundPtr >= numAlloc) {
 		numAlloc += EXTRA_ENTRIES;
-		*foundPtrPtr =(int*)ckrealloc(*foundPtrPtr,numAlloc*sizeof(int));
+		*foundPtrPtr =(int*)ckrealloc(*foundPtrPtr,
+					      numAlloc*sizeof(int));
 	    }
 	    (*foundPtrPtr)[(*numFoundPtr)++] = i;
 	}
     }
 
-    ckfree((char*) expWords);
     ckfree((char*) notPtr);
     ckfree((char*) fieldPtr);
     ckfree((char*) valuePtr);
@@ -1374,7 +1381,7 @@ RatDbGetMessage(Tcl_Interp *interp, int index, char **buffer)
 
     Unlock(interp);
 
-    return RatParseMsg(interp, message);
+    return RatParseMsg(interp, (unsigned char*)message);
 }
 
 
@@ -1756,16 +1763,11 @@ RatDbDaysSinceExpire(Tcl_Interp *interp)
      * First make sure we know where the database should reside.
      */
     if (0 == dbDir) {
-	Tcl_DString ds;	/* Scratch area */
-	char *value;	/* Value of "dbase_dir" option */
-
-	value = Tcl_GetVar2(interp, "option", "dbase_dir", TCL_GLOBAL_ONLY);
+	const char *value = RatGetPathOption(interp, "dbase_dir");
 	if (NULL == value) {
-	    Tcl_SetResult(interp, "option(dbase_dir) has no value", TCL_STATIC);
 	    return TCL_ERROR;
 	}
-	dbDir = cpystr(Tcl_TildeSubst(interp, value, &ds));
-	Tcl_DStringFree(&ds);
+	dbDir = cpystr(value);
     }
 
     snprintf(buf, sizeof(buf), "%s/expired", dbDir);
@@ -1805,7 +1807,8 @@ RatDbDaysSinceExpire(Tcl_Interp *interp)
 int
 RatDbExpire(Tcl_Interp *interp, char *infolder, char *backupDirectory)
 {
-    char *compressProg, *compressSuffix, *backupDir, *statusId;
+    char *compressProg, *compressSuffix, *statusId;
+    const char *backupDir;
     int numScan = 0, numDelete = 0, numBackup = 0, numInbox = 0, numCustom = 0;
     int i, len, delete, mode, fd, doBackup = 0, changed = 0, error = 0;
     int move;
@@ -1826,8 +1829,8 @@ RatDbExpire(Tcl_Interp *interp, char *infolder, char *backupDirectory)
     /*
      * Get file creation mode
      */
-    Tcl_GetInt(interp, Tcl_GetVar2(interp, "option", "permissions",
-	    TCL_GLOBAL_ONLY), &mode);
+    oPtr = Tcl_GetVar2Ex(interp, "option", "permissions", TCL_GLOBAL_ONLY);
+    Tcl_GetIntFromObj(interp, oPtr, &mode);
     /*
      * Make sure the inbox directory exists.
      */
@@ -1852,7 +1855,7 @@ RatDbExpire(Tcl_Interp *interp, char *infolder, char *backupDirectory)
     }
     compressProg = getenv("COMPRESS");
     compressSuffix = getenv("CSUFFIX");
-    backupDir = Tcl_GetVar2(interp, "option", "dbase_backup", TCL_GLOBAL_ONLY);
+    backupDir = RatGetPathOption(interp, "dbase_backup");
     if (!compressProg || !compressSuffix || !backupDir) {
 	Tcl_AppendResult(interp, "Internal error: compressProg, ",
 		"compressSuffix or option(dbase_backup) not defined",
@@ -1967,11 +1970,11 @@ RatDbExpire(Tcl_Interp *interp, char *infolder, char *backupDirectory)
      */
     if (numBackup) {
 	int chunkSize;
-	char *chunk;
+	Tcl_Obj *oPtr;
 
-	chunk = Tcl_GetVar2(interp, "option", "chunksize", TCL_GLOBAL_ONLY);
-	if (chunk) {
-	    Tcl_GetInt(interp, chunk, &chunkSize);
+	oPtr = Tcl_GetVar2Ex(interp, "option", "chunksize", TCL_GLOBAL_ONLY);
+	if (oPtr) {
+	    Tcl_GetIntFromObj(interp, oPtr, &chunkSize);
 	} else {
 	    chunkSize = 100;
 	}
@@ -1992,7 +1995,7 @@ RatDbExpire(Tcl_Interp *interp, char *infolder, char *backupDirectory)
     }
     if (doBackup) {
 	Tcl_Channel backupChannel, inChannel;
-	char *argv[3], *error = NULL;
+	CONST84 char *argv[3], *error = NULL;
 
 	ckfree(statusId);
 	RatLogF(interp, RAT_INFO, "packing_backup", RATLOG_EXPLICIT);
@@ -2031,7 +2034,7 @@ RatDbExpire(Tcl_Interp *interp, char *infolder, char *backupDirectory)
 	    }
 	    if (error) {
 		unlink(buf2);
-		Tcl_SetResult(interp, error, TCL_STATIC);
+		Tcl_SetResult(interp, (char*)error, TCL_STATIC);
 		Tcl_BackgroundError(interp);
 	    } else {
 		rewinddir(dirPtr);
@@ -2250,7 +2253,7 @@ RatDbBuildList(Tcl_Interp *interp, Tcl_DString *dsPtr, char *prefix, char *dir,
 	    if (*prefix) {
 		snprintf(buf, sizeof(buf), "%s/%s", prefix, entPtr->d_name);
 	    } else {
-		RatStrNCpy(buf, entPtr->d_name, sizeof(buf));
+		strlcpy(buf, entPtr->d_name, sizeof(buf));
 	    }
 	    itemPtr = (RatDbItem*)ckalloc(sizeof(RatDbItem));
 	    itemPtr->fileSize = sbuf.st_size;
@@ -2264,7 +2267,7 @@ RatDbBuildList(Tcl_Interp *interp, Tcl_DString *dsPtr, char *prefix, char *dir,
 	    if (prefix && *prefix) {
 		snprintf(path, sizeof(path), "%s/%s", prefix, entPtr->d_name);
 	    } else {
-		RatStrNCpy(path, entPtr->d_name, sizeof(path));
+		strlcpy(path, entPtr->d_name, sizeof(path));
 	    }
 	    snprintf(buf, sizeof(buf), "%s/%s", dir, entPtr->d_name);
 	    RatDbBuildList(interp, dsPtr, path, buf, tablePtr, fix);
@@ -2301,11 +2304,11 @@ int
 RatDbCheck(Tcl_Interp *interp, int fix)
 {
     int numFound = 0, numMal = 0, numAlone = 0, numUnlinked = 0, totSize = 0,
-        fd, lines, start, index, i, j, extraNum = 0, extraAlloc = 0, msgLen = 0,
-	date = 0, listArgc, elemArgc, indexInfo = 0, numDel = 0;
+        fd, lines, start, index, i, j, extraNum = 0, extraAlloc = 0,
+	msgLen = 0, date = 0, listArgc, elemArgc, indexInfo = 0, numDel = 0;
     char buf[8092], *indexPtr = NULL, **linePtrPtr = NULL, *cPtr,
-	 *to, *from, *cc, *subject, *flags, **listArgv, **elemArgv,
-	 *msgBuf = NULL;
+	 *to, *from, *cc, *subject, *flags, *msgBuf = NULL;
+    CONST84 char **listArgv, **elemArgv;
     Tcl_HashTable items, status;
     char **extraPtrPtr = NULL;
     Tcl_HashEntry *entryPtr;
@@ -2321,15 +2324,11 @@ RatDbCheck(Tcl_Interp *interp, int fix)
      * Initialize variables
      */
     if (0 == dbDir) {
-	Tcl_DString ds;	/* Scratch area */
-	char *value;	/* Value of "dbase_dir" option */
-
-	value = Tcl_GetVar2(interp, "option", "dbase_dir", TCL_GLOBAL_ONLY);
+	const char *value = RatGetPathOption(interp, "dbase_dir");
 	if (NULL == value) {
-	    return TCL_OK;
+	    return TCL_ERROR;
 	}
-	dbDir = cpystr(Tcl_TildeSubst(interp, value, &ds));
-	Tcl_DStringFree(&ds);
+	dbDir = cpystr(value);
     }
     if (0 == ident) {
 	gethostname(buf, sizeof(buf));
@@ -2527,7 +2526,8 @@ RatDbCheck(Tcl_Interp *interp, int fix)
 	    for (i=0; i<RATDBETYPE_END; i++, start++) {
 		if (i == STATUS && (entryPtr = Tcl_FindHashEntry(&status,
 			(char*)numFound))) {
-		    itemPtr->entry.content[i]=(char*)Tcl_GetHashValue(entryPtr);
+		    itemPtr->entry.content[i] =
+			(char*)Tcl_GetHashValue(entryPtr);
 		} else {
 		    itemPtr->entry.content[i] = linePtrPtr[start];
 		}
@@ -2573,7 +2573,9 @@ RatDbCheck(Tcl_Interp *interp, int fix)
 	    msgBuf[itemPtr->fileSize] = '\0';
 	    close(fd);
 	    if (NULL == (cPtr = strstr(msgBuf, "\n\n"))) {
-		cPtr = strstr(msgBuf, "\r\n\r");
+		if (NULL == (cPtr = strstr(msgBuf, "\r\n\r"))) {
+		    cPtr = msgBuf + strlen(msgBuf);
+		}
 	    }
 	    *(++cPtr) = '\0';
 	    RatMessageGetHeader(interp, msgBuf);
@@ -2599,7 +2601,7 @@ RatDbCheck(Tcl_Interp *interp, int fix)
 			flags = cpystr(elemArgv[1]);
 		    }
 		} else if (!strcasecmp(elemArgv[0], "date")) {
-		    if (T == mail_parse_date(&elt, elemArgv[1])) {
+		    if (T == mail_parse_date(&elt, (char*)elemArgv[1])) {
 			tm.tm_sec = elt.seconds;
 			tm.tm_min = elt.minutes;
 			tm.tm_hour = elt.hours;
@@ -2625,15 +2627,16 @@ RatDbCheck(Tcl_Interp *interp, int fix)
 	    itemPtr->entry.content[CC] = cc ? cc : "";
 	    itemPtr->entry.content[SUBJECT] = subject ? subject : "";
 	    sprintf(buf, "%d", date);
-	    itemPtr->entry.content[DATE] = extraPtrPtr[extraNum++] =cpystr(buf);
+	    itemPtr->entry.content[DATE] = extraPtrPtr[extraNum++]=cpystr(buf);
 	    itemPtr->entry.content[KEYWORDS] = "LostMessage";
 	    sprintf(buf, "%d", itemPtr->fileSize);
-	    itemPtr->entry.content[RSIZE] = extraPtrPtr[extraNum++]=cpystr(buf);
+	    itemPtr->entry.content[RSIZE] =extraPtrPtr[extraNum++]=cpystr(buf);
 	    itemPtr->entry.content[STATUS] = flags ? flags : "";
 	    sprintf(buf, "%ld", time(NULL)+60*60*24*100);
-	    itemPtr->entry.content[EX_TIME]=extraPtrPtr[extraNum++]=cpystr(buf);
+	    itemPtr->entry.content[EX_TIME] = extraPtrPtr[extraNum++] =
+		cpystr(buf);
 	    itemPtr->entry.content[EX_TYPE] = "backup";
-	    itemPtr->entry.content[FILENAME] = Tcl_GetHashKey(&items, entryPtr);
+	    itemPtr->entry.content[FILENAME] = Tcl_GetHashKey(&items,entryPtr);
 	}
     }
     if (numUnlinked && fix) {
@@ -2642,10 +2645,12 @@ RatDbCheck(Tcl_Interp *interp, int fix)
     }
     if (indexInfo != numFound+numUnlinked-numDel) {
 	if (fix) {
-	    sprintf(buf, "Number of messages in index.info was wrong (was: %d is now: %d)",
+	    sprintf(buf, "Number of messages in index.info was wrong "
+		    "(was: %d is now: %d)",
 		indexInfo, numFound+numUnlinked-numDel);
 	} else {
-	    sprintf(buf, "Number of messages in index.info is wrong (was: %d should be: %d)",
+	    sprintf(buf, "Number of messages in index.info is wrong "
+		    "(was: %d should be: %d)",
 		indexInfo, numFound+numUnlinked-numDel);
 	}
 	Tcl_DStringAppendElement(&reportDS, buf);
@@ -2680,7 +2685,7 @@ RatDbCheck(Tcl_Interp *interp, int fix)
 
 	if (isRead) {
 	    isRead = 0;
-	    sprintf(buf, "Popup $t(need_restart)");
+	    strlcpy(buf, "Popup $t(need_restart)", sizeof(buf));
 	    Tcl_Eval(interp, buf);
 	}
     }
@@ -2697,12 +2702,8 @@ RatDbCheck(Tcl_Interp *interp, int fix)
     }
     Tcl_DeleteHashTable(&items);
     Tcl_DeleteHashTable(&status);
-    if (indexPtr) {
-	ckfree(indexPtr);
-    }
-    if (linePtrPtr) {
-	ckfree(linePtrPtr);
-    }
+    ckfree(indexPtr);
+    ckfree(linePtrPtr);
 
     Tcl_ResetResult(interp);
     sprintf(buf, "%d", numFound);
@@ -2745,7 +2746,7 @@ RatDbCheck(Tcl_Interp *interp, int fix)
  */
 
 static int
-NoLFPrint(FILE *fp, char *s)
+NoLFPrint(FILE *fp, const char *s)
 {
     unsigned char *cPtr;
 
