@@ -40,6 +40,13 @@
 #include <X11/Xutil.h>
 #include <X11/Xatom.h>
 
+#include <stdlib.h>
+#include "blt.h"
+
+#ifndef CONST84
+#   define CONST84
+#endif
+
 #define TRUE    1
 #define FALSE   0
 
@@ -1080,7 +1087,7 @@ BusyCmd(clientData, interp, argc, argv)
     ClientData clientData;	/* Main window of the interpreter */
     Tcl_Interp *interp;		/* Interpreter associated with command */
     int argc;
-    char **argv;
+    CONST84 char **argv;
 {
     int result;
 
